"""Documentation/tests for the Walking Deque Class

A deque with custom rotation and memory operations.

### Preliminary imports

Since some of our tests/examples use some standard library
imports, let's go ahead and perform them:

>>> from collections import deque
>>> from random      import seed, randint
>>> from copy        import deepcopy

Let's also import the class we are testing
>>> from common.wdeque.main import WalkingDeque

### Purpose and general usage

This class was first designed to assist in game animation.
Basically it is a good solution for any problem which demands
quick switching between a finite set of values many times
over.

In that respect, it is specially useful for animation,
because it can be used to quickly switch between animation
values like surfaces (images) and positions in 2d space
(containers of x, y coordinates representing 2d points).

To be more precise, we use the walking deque to switch
between indices which point to these values in other
containers like lists.

>>> # animation values (positions):
>>> position_list = [(0, 0), (5, 0), (10, 0), (15, 0)]
>>> # indices:
>>> walking_deque = WalkingDeque([0, 0, 1, 1, 2, 3])

>>> # loop:
>>> while True:
...     index = walking_deque[0]
...     print(position_list[index])
...     walking_deque.walk(1)
...     if walking_deque.loops_no == 2.0: break
... 
(0, 0)
(0, 0)
(5, 0)
(5, 0)
(10, 0)
(15, 0)
(0, 0)
(0, 0)
(5, 0)
(5, 0)
(10, 0)
(15, 0)

The while loop above illustrates the animation process. It
could be said that each index represents a frame of
animation and is used to retrieve a specific animation value.
Each loop the deque walks and the value in its first index
is retrieved and used as the index to retrieve the position
value in the position_list.

Additionally, we added a condition to break out of the while
loop when the deque reaches 2 complete loops around its own
content.

Because lots of problems can be generalized as an animation
problem, this class usage has grown beyond of its animation
"role". For instance, I also use it to switch between GUI
items by binding the "walk" method to keys so the user can
walk the different GUI items instead of having them
automatically walked as in the example above illustrating an
animation.

In summary, any future changes on this class and its
dependencies must be done with the goal of keeping the
efficiency of the switching process in mind, mostly for
animation but also for other more general purposes related
to switching between finite sets of values.


### Introduction

This deque uses a different nomenclature for rotation which
is called "walk". Walking means rotating the deque so that
the items closer to the end get closer to the beginning. This
means that, apart from other extra administrative tasks the
walk operation have, the walk is the opposite operation of
the regular deque rotation. As you can observe:

>>> items = [0, 1, 2, 3]

>>> # though both deques will perform rotation operations
>>> # with 1, the effects will be opposite:

>>> regular_deque = deque(items)
>>> regular_deque.rotate(1)
>>> print(items, regular_deque)
[0, 1, 2, 3] deque([3, 0, 1, 2])

>>> walking_deque = WalkingDeque(items)
>>> walking_deque.walk(1)
>>> print(items, walking_deque)
[0, 1, 2, 3] WalkingDeque([1, 2, 3, 0])

The "walk" analogy works perfectly when we think of walking
in real life, where things which are farther begin to get
closer as we walk further. In the example above, for instance,
by walking one step, the walking deque first index went from
0 to 1. By rotating 1 step, on the other hand, the regular
deque first index went from 0 to 3, a much more radical
difference.

We could just use the "rotate" nomenclature and invert its
signal in this class, but we believe different things must be
named differently, to avoid confusion. Thus, we created a new
method called walk.


### Memory operations

Apart from its different rotation operation, the walking
deque also has memory operations and related attributes. For
instance, it accumulates the number of steps performed in the
"total_walking" attribute, regardless of the direction used.

>>> items = [0, 1, 2, 3]
>>> w = WalkingDeque(items)
>>> w.walk( 3) # walk    3 steps
>>> w.walk(-1) # go back 1 step
>>> w.total_walking
2

You can also see the total number of loops performed

>>> # the result of the operation above is 0.5 because
>>> # it is of length 4 and walked 2 steps in total
>>> w.loops_no
0.5

Notice that both the total_walking and loops_no attributes
would be holding negative numbers now if we used regular
rotation, but we are -- advancing -- through the items in the
deque, so again, for our purposes, "walk" is a much more
intuitive operation, which uses positive steps to advance
and negative steps to go back.

The loops_no is actually the result of dividing the total
walking by the length of the deque

>>> w.total_walking / len(w)
0.5
>>> w.total_walking / len(w) == w.loops_no
True

As we discuss further ahead, the length of a walking deque
can't be changed, so the length is stored in the 'length'
attribute for speed, since acessing the attribute is slightly
faster then using the builtin "len".

>>> w.total_walking / w.length == w.loops_no
True

Remember that all attributes above have their signals changed
when the walking is performed with negative steps.

>>> items = [0, 1, 2, 3]
>>> w = WalkingDeque(items)
>>> w.walk(-1) # walk 1 steps back
>>> w.walk(-1) # walk 1 steps back
>>> w.total_walking
-2
>>> w.loops_no # we walk half a loop backwads
-0.5

Also, since it remembers the steps, it can revert back to its
initial state. That is, no matter how many times you walk and
in which diretion, by calling the "restore_walking" method
the deque returns to its original order, when it was
instantiated.

>>> seed()
>>> w = WalkingDeque(range(4))
>>> w
WalkingDeque([0, 1, 2, 3])
>>> # let's now walk a random amount of steps
>>> seed()
>>> w.walk(randint(-3, 3))

>>> # now notice that, no matter the steps walked,
>>> # we'll really go back to the starting position
>>> w.restore_walking()
>>> w
WalkingDeque([0, 1, 2, 3])

As regular deques, indexing works fine and slicing isn't
allowed:

>>> w[0]
0
>>> w[-1]
3
>>> w[1:3]
Traceback (most recent call last):
...
TypeError: sequence index must be integer, not 'slice'

In addition to getting items in their current indexes by
using the usual self[index] syntax as seen above, you can
also pass any integer to the get_mem_value method and it will
return the value for that position as if the deque had never
walked.

>>> w = WalkingDeque([11, 22, 33, 44])
>>> w
WalkingDeque([11, 22, 33, 44])
>>> w.walk(2)
>>> w
WalkingDeque([33, 44, 11, 22])
>>> # give me the value of 0 when deque was instantiated:
>>> w.get_mem_value(0)
11
>>> # give me the value of -2 when deque was instantiated:
>>> w.get_mem_value(-2)
33

>>> # now let's walk a random amount and see if get_mem_value
>>> # still works properly
>>> w.walk(randint(1, 3))
>>> w.get_mem_value(0)
11
>>> w.get_mem_value(-2)
33
>>> # yes, it works fine!

>>> # however, the index pased must be within range:
>>> w.get_mem_value(-5)
Traceback (most recent call last):
...
IndexError: deque index out of range

Additionaly, the 'get_unwalked_list' method will return a
list of all the items in their original positions.

>>> w.get_unwalked_list()
[11, 22, 33, 44]

>>> # notice that passing an index to get_mem_value or
>>> # trying to get it from the returned unwalked list
>>> # produces the same result:

>>> # walk a random amount of steps just to illustrate
>>> # how this operations is totally independent of the
>>> # walked steps
>>> w.walk(randint(-3, 3)) 
>>> # now the proof:
>>> w.get_mem_value(-3) == w.get_unwalked_list()[-3]
True

Another useful method of the walking deque is the
get_original_index. The get_original_index works similarly to
the get_mem_value method, but instead of passing a index to
know which value was sitting there when the deque had not
walked, you pass the current index of a value in the deque
and the method will return the index where the value was
located when the deque was instantiated, that is, before any
walking was ever performed.

For instance:

>>> w = WalkingDeque([11, 22, 33, 44])
>>> w.walk(2)
>>> w[0] # 33 is currently on the 0 (zero-eth) index
33
>>> # but in which index was 33, before any walking was ever
>>> # performed? Easy: we just pass its current index, which
>>> # is 0 (zero), to the get_original_index method:
>>> w.get_original_index(0)
2
>>> # now we now the original index of the item in 0 (33)
>>> # was 2.

More examples:

>>> w = WalkingDeque([11, 22, 33, 44])
>>> w.walk(2)
>>> w.get_original_index(0)
2
>>> w.get_original_index(1)
3
>>> # negative indices work too:
>>> w.get_original_index(-1)
1
>>> # we have a functools.partialmethod for
>>> # "get_original_index(0)", since we used it a lot in
>>> # animation
>>> w.get_index_of_first()
2
>>> # the index must be within range, though:
>>> w.get_original_index(4)
Traceback (most recent call last):
...
IndexError: deque index out of range

The walking deque class also has the peek_loops_no method,
which returns the loops_no value as if you had walked a given
number of steps. This is useful to find out how much more
steps it will take to complete a certain number of loops.

For instance:

>>> w = WalkingDeque([0, 1, 2, 3])
>>> w.walk(3)
>>> # the result below means in one more step you'll
>>> # reach 1 complete loop around its content
>>> w.peek_loops_no(1)
1.0
>>> # the result below on the other hand means that if
>>> # I go one step back I'll be at 0.5 loops around
>>> # the deque's content
>>> w.peek_loops_no(-1)
0.5

This is specially useful when you need to do additional
setups before the walking deque reaches a given number of
loops.


### Empty instantiation not allowed

The walking deque must also have its items fed to it upon
instantiation. It doesn't allow length altering operations
(more about it in the 'Immutable length' discussion ahead).
Thus, you must not ommit an iterable upon instantiation, nor
must you pass an empty one:

>>> # no iterable was passed:
>>> w = WalkingDeque() # doctest: +ELLIPSIS
Traceback (most recent call last):
...
TypeError: __init__() missing 1 required positional...
>>> # the error raised above default behaviour triggered by
>>> # not complying with the collections.deque.__init__
>>> # method signature

>>> # a value error is also triggered if the iterable
>>> # passed is empty:
>>> items = []
>>> w = WalkingDeque(items)
Traceback (most recent call last):
...
ValueError: iterable must not be empty


### Item reassignment

Item reassignment is allowed though. The deque will treat the
new item as if it were the old one, that is, the rotation
information will remain unchanged (the values of loops_no and
total_walking attributes). This causes no harm nor confusion
regarding the problems this class was designed to handle.

Using an analogy, you can think of item reassignement as an
exchange of "passengers". The old passenger is replaced, but
the mileage of the vehicle is the same.

>>> # let's create a deque, walk it a bit, and display
>>> # it's memory related data
>>> w = WalkingDeque([0, 1, 2, 3])
>>> w.walk(3)
>>> w.walk(-1)

>>> w
WalkingDeque([2, 3, 0, 1])
>>> w.total_walking
2
>>> w.loops_no
0.5

>>> # now notice that even though we reassigned a value,
>>> # the walking related data stays the same
>>> w[0] = 7
>>> w
WalkingDeque([7, 3, 0, 1])
>>> w.total_walking
2
>>> w.loops_no
0.5

Remember that, as we discussed before, the default deque
rotation operation (collections.deque.rotate) was
suppressed. So, it can't be used:

>>> w = WalkingDeque(range(4))
>>> w.rotate(1)
Traceback (most recent call last):
...
RuntimeError: operation ruled out by the design

Ahead in the documentation you'll learn about more methods
which were suppressed and thus raise the same error.

### Content views

Another ability of the walking deque is to generate special
proxy-like objects called "content views". They work like
read-only copies/proxies of the walking deques which
originated them.

Content views have their walking independent of the walking
of their original deque, but their contents are linked to
the contents in their "proxied" deque, which means every
time you change a value in the proxied deque the
corresponding value changes in the view. Such corresponding
value is the value which sits in the same original index.

Notice, below, that the content view create isn't affected
by the walking of the walking deque which generated it:

>>> w = WalkingDeque(range(4))
>>> v = w.get_content_view()
>>> w.walk(2)
>>> w
WalkingDeque([2, 3, 0, 1])
>>> v
ContentView([0, 1, 2, 3])

When we change the value of the index 0 of the original
walking deque, the index 2 of the view is changed
accordingly:

>>> w[0] = 45
>>> w
WalkingDeque([45, 3, 0, 1])
>>> v
ContentView([0, 1, 45, 3])

This is so because both they are corresponding values, that
is, they were in the same position when the deques were
created. In other words, the original index of w[0] and v[0]
are the same:

>>> w.get_original_index(0) == v.get_original_index(2)
True

See how their items are aligned when "unwalked":
>>> print(w.get_unwalked_list(), v.get_unwalked_list())
[0, 1, 45, 3] [0, 1, 45, 3]

## Reassigning indices in content views

Content views are protected from external changes. Notice you
can't assign values in the view:

>>> v[0] = 13
Traceback (most recent call last):
...
ValueError: value must be identical to proxied deque value

However, there's a way to do that: if you assign a value
which is the same value as the one present in the proxied
deque's corresponding original index, then the operation is
considered legal. Though in this cases it changes nothing,
it is also a harmless operation.

>>> v[0]
0
>>> v[0] = 0
>>> # notice that nothing changed, since v[0] already held
>>> # the value 0 (zero) before we reassigned it.
>>> # Nonetheless, the operation didn't raise any errors.

Though it is a seemingly useless operation here, it is
crucial for the automatic update process between walking
deques and their respective view instances.

The custom item reassignment implemented in the walking
deque class, which sits at the __setitem__ method, first
changes the value in its own item, then proceed to change
the value in its views, if the deque has any view.

Normally, reassigning values on the views would raise errors,
but since the new values are now also present in the proxied
deque, the operation is legal. See the
ContentView.__setitem__ to learn how this works exactly.

Notice, however, that it's not enough for the objects/values
to be equal: they must be --the same-- object, which means
they must pass an identity check (a is b):

>>> d = dict(hello='world')
>>> w = WalkingDeque([0, 1, 2, d])
>>> c = w.get_content_view()

>>> d_copy = d.copy() # generate an equal dict
>>> d == d_copy
True
>>> # this will cause an error even though, as we saw above
>>> # the dicts have the same value
>>> c[3] = d_copy
Traceback (most recent call last):
...
ValueError: value must be identical to proxied deque value
>>> # This is ok
>>> c[3] = d


## self.views attribute

All the content views generated by a walking deque are also
automatically stored in a list in the views attribute:

>>> w.views
[ContentView([0, 1, 2, {'hello': 'world'}])]

For more information and examples, check the docstring in the
ContentView class.


### Immutable length

Finally, as previously stated, this class was designed from
the premisse that it would receive its complete set of values
when instantiated.

No thought was spared about its behaviour in the case it was
built little by little by appending values or in the case it
started with values but would have some of those values poped
out.

It doesn't mean the design process was lax. On the contrary,
the methods were built and refined little by little based on
real problems related to its main purpose of serving as a
value switcher for game animation values. It is just that
such cases were never necessary, even for the most complex
problems.

However, out of sheer curiosity, I tried to come up with a
design which would allow the existence of those cases.
This was also done as a preventive measure, since the need
for such "feature" might come up one day and it would be
good to have the design at least sketched out for it.

This turned out to be a far complex task than I anticipated,
though. Suppose, for instance, that you have a walking deque
with some values which walked a number of times. In case you
want to remove an item, should you reset the memory related
attributes when doing so?

In case you want to add a new value, should you restore the
walking, add the value and just pretend the deque was just
created? How to recalculate the loops_no attribute in such
cases?

And the total_walking? What should the get_unwalked_list
method return then? Should you restore the original unwalked
state, add the item and them walk back the total walking or
walk back until the loops_no is the same or roughly the same?

As you can see, many questions pop up when trying to
implement that and the solutions often have lots of trade
offs. Thus, I decided to visit this problem again only when
I truly need said feature. I won't deny this is a very
interesting problem, though.

Most importatly, it is a fact, however, that none of the
use cases I handle --need-- said feature, including problems
related to game animation which is the main purpose of this
class. Hence it would be difficult to assess the efficacy of
possible solutions without any context/use case on which to
test it. Thus, even if I were to come up with a solution, it
would remain recorded on paper until the day came when it
would be implemented.

In conclusion, having no time to spare due to the amount of
work to be done, I chose to leave this "design challenge"
aside for now. I call it a design challenge since it isn't
something we need, so it would be inaproppriate to say the
absence of this specific design is a problem per se. 


### Consequences of the immutable length design

Since instances of this class must not have their lengths
altered, all method which do so were suppressed and won't work,
as you can see:

>>> w = WalkingDeque(range(4))

>>> w.append(1)
Traceback (most recent call last):
...
RuntimeError: operation ruled out by the design
>>> w.extend([11, 22, 33])
Traceback (most recent call last):
...
RuntimeError: operation ruled out by the design
>>> del w[0] # __delitem__ method
Traceback (most recent call last):
...
RuntimeError: operation ruled out by the design

All suppressed methods were: append, appendleft, extend,
extendleft, pop, popleft, rotate, insert, clear, remove
and __delitem__ and the ones describe in the next subsection.


## Additional suppressed methods

Other methods, though they don't alter the length of the
deque were also suppressed to prevent errors since they don't
serve any purpose in this class and/or have effects which
prevent the correct usage of the class. Those methods were
"reverse" and "__delattr__":

>>> w = WalkingDeque(range(4))
>>> w.reverse()
Traceback (most recent call last):
...
RuntimeError: operation ruled out by the design
>>> del w.loops_no # __delattr__
Traceback (most recent call last):
...
RuntimeError: operation ruled out by the design


## suppressing behaviour

The suppressing behaviour was implemented by creating a
method with general signature (*args and **kwargs) called
"_suppressed" which simply raises an error explaining the
suppression. This method is then used to override all the
methods we want to suppress by assigning _suppressed to their
names. Thus, every suppressed method name now stores the
_suppressed method:

>>> WalkingDeque.rotate == WalkingDeque._suppressed
True


### Custom deepcopy operation

A custom deepcopy operation was implemented in the
__deepcopy__ method, as instructed by the standard library
copy module official documentation (for python version 3.5).

It is needed for two reasons:

1) Besides the walking deque data, we also generate
   content view objects (special copies/proxies) which
   are referenced in the walking deque and which contain
   references to the walking deque which generated them.
   Because of that, the content views also have their own
   custom deepcopy operation to guarantee that the
   proxy-view relationship is preserved between deepcopies
   of those objects when they are deepcopied recursively.

2) We ommit the maxlen parameter from the signature of
   the collections.deque.__init__ method, which also causes
   problems when using the inherited __deepcopy__ operation,
   which also makes a custom deepcopy operation needed.

>>> w = WalkingDeque(range(4))
>>> c = deepcopy(w)
>>> w.get_unwalked_list() == c.get_unwalked_list()
True
>>> w is c
False

When deepcopied, walking deques are often nested within
dictionaries. In some cases such dictionaries will also have
references to their respective content views. For instance:

>>> w  = WalkingDeque([0, 1, 2, 3])
>>> v1 = w.get_content_view()
>>> v2 = w.get_content_view()
>>> d = {
...   "original": w,
...   "view1"   : v1,
...   "view2"   : v2
... }
>>> new_dict = deepcopy(d)
>>> new_w    = new_dict["original"]
>>> new_v1   = new_dict["view1"]
>>> new_v2   = new_dict["view2"]
>>> new_v1.proxied is new_w
True
>>> new_v2.proxied is new_w
True
>>> new_w.walk(2)
>>> new_w[0] = 34
>>> new_v1[2] == 34
True

Notice that by deepcopying the dict, all objects inside were
deepcopied recursively which means no object was deepcopied
more than once. This is how deepcopying naturally works in
recursive deepcopy operations.

This kind of deepcopy, called recursive, which avoids
deepcopying objects more than once and keep their references
and therefore their relationships intact, only happens when
you deepcopy an object which has other objects embedded or
contained in it (via attributes, indices or keys, for
instance).

Unless, of course, the object in question is a custom object
with a custom deepcopy implementation, then you'll have to
check if the deepcopy operation performed behaves that way.

When you deepcopy objects one by one, though, then the
deepcopies generated won't keep their relationships. See:

>>> # creating walking deque and its view
>>> orig  = WalkingDeque([0, 1, 2, 3])
>>> view = orig.get_content_view()

>>> # deepcopying them one at a time
>>> new_orig = deepcopy(orig)
>>> new_view = deepcopy(view)

>>> # notice how the new_view proxied deque isn't new_orig:
>>> new_view.proxied is new_orig
False
>>> # the new view also isn't referenced in the new_orig
>>> # views attribute (a list):
>>> new_orig.views[0] is new_view
False

This means new_view lost its relationship with new_orig and
won't respond to changes on new_orig:

>>> new_orig[0] = 99
>>> new_view[0] == 99
False

Therefore, when you desire to keep the relationship between
the deepcopies, make sure to deepcopy them recursively, that
is, by storing references to them in the same objects. For
instance, deepcopy them inside a dictionary as we did before
or maybe inside a tuple:

>>> w  = WalkingDeque([0, 1, 2, 3])
>>> v1 = w.get_content_view()

>>> # deepcopy them inside a tuple, for instance
>>> new_w, new_v1 = deepcopy((w, v1))

>>> # now they kept their relationship:
>>> new_v1.proxied is new_w
True
>>> new_w[0] = 144
>>> new_v1[0] == 144
True
>>> new_w.views[0] is new_v1
True

### Extra notes:


## maxlen parameter absence in __init__ method

Tests I performed showed that passing or not passing a maxlen
argument when creating deques don't have practically any
impact in rotation operation speed. Sometimes one case wins,
sometimes the other, but even so with roughly the same timing
(the difference almost reach the nanoseconds mark), indicating
that the presence or absence of the maxlen parameter isn't a
significant factor for speed.

Since walking deques aren't meant to have their length
altered though, the parameter was removed from the
constructor method (__init__). It's presence just doesn't
make sense and represents an useless additional parameter
about which to worry, so its removal is only natural.

As already explained, such removal was one of the causes of
the implementation of a custom deepcopy behaviour.


## About length, loops_no and total_walking attributes

I could protect those attributes against thoughtless meddling
by providing a setter for each of them or by implementing a
__setattr__ method with extra checks, but since this is a
class designed for efficiency, the extra overhead is an
undesirable trade off, thus such possibility was ignored.
At least until we have a better idea.
"""



"""Facility for GameLevel class tools."""

from pygame import Rect

from ..config import SCREEN, SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_RECT
from ..appcommon.surf import render_rect
from ..appcommon.autoblit import BlitterSet
from ..palette import WHITE


class ScrollBarrier(object):
    """A barrier to trigger scrolling when touching player."""

    def __init__(self, rect, scroll_command):
        """Initialize superclass and perform setup."""
        self.invoke = scroll_command

        self.rect = rect
        self.image = render_rect(*self.rect.size, color=WHITE)

    def update(self):
        """Update object state."""
        SCREEN.blit(self.image, self.rect)


def set_scroll_handles(level):
    """Set level handles for scrolling.

    level
        Instance of gamelevel.main.Level.
    """
    vertical_handle_width = horiz_handle_height = 10

    ## left
    left_rect = Rect((0, 0), (vertical_handle_width, SCREEN_HEIGHT))
    left_rect.midleft = SCREEN_RECT.midleft
    left_barrier = ScrollBarrier(left_rect, level.scroll_left)

    ## right
    right_rect = Rect((0, 0), (vertical_handle_width, SCREEN_HEIGHT))
    right_rect.midright = SCREEN_RECT.midright
    right_barrier = ScrollBarrier(right_rect, level.scroll_right)

    ## upper
    upper_rect = Rect((0, 0), (SCREEN_WIDTH, horiz_handle_height))
    upper_rect.midtop = SCREEN_RECT.midtop
    upper_barrier = ScrollBarrier(upper_rect, level.scroll_up)

    ## lower
    lower_rect = Rect((0, 0), (SCREEN_WIDTH, horiz_handle_height))
    lower_rect.midbottom = SCREEN_RECT.midbottom
    lower_barrier = ScrollBarrier(lower_rect, level.scroll_down)

    ## Setups
    level.scroll_barriers = BlitterSet(
        [left_barrier, right_barrier, upper_barrier, lower_barrier]
    )

    level.horizontal_barriers = BlitterSet([left_barrier, right_barrier])

    level.vertical_barriers = BlitterSet([upper_barrier, lower_barrier])



"""Facility for single line updateable text object."""

from string import ascii_letters, digits, punctuation
from random import seed, choice

### third-party import
from pygame import Rect, error as pygame_error

### local imports

from ....common.behaviour import empty_function

from ...surf import render_rect
from ...autoblit import BasicObject
from ..main import render_text

from ....font import CHAR_CODE_MAP
from ....palette import BLACK

# TODO implement recent changes here and in the autolabel
# regarding the custom coordinates arguments and how they
# are automatically set in the __init__ and set methods
# of this label class in the respective classes from the
# animation editor; remember to take into account the
# auto position correction on update image when the labels
# are offset by hide switch objects (on the animation editor
# package)

seed()

###### constant definitions

CHAR_DATA_MAP = {}

COMMON_CHARS = ascii_letters + digits + punctuation + " " + "áàãâêéèẽîíìôóòõûúùñç"

##### utility functions


def safe_text_render(char, render_kwargs):
    """Render char with extra except clause for safety."""
    try:
        return render_text(char, padding=0, **render_kwargs)

    ### errors raised by pygame.font.Font.render:
    ### some values aren't accepted in pygame (they are
    ### called null values) and raise value errors (until
    ### now, only chr(0) caused this, maybe it's the only
    ### one?);
    ### a pygame error sometimes is also raised saying
    ### "Text has zero width" (until now, only chr(13)
    ### caused this, maybe it's the only one?);
    ### those problems vary depending on the font; such
    ### errors where actually found in another codebase;
    except (ValueError, pygame_error):
        pass


# XXX consider a better name for this function; it is
# difficult because it does many things, but at the same
# time I don't think the current name is appropriate;
def create_char_maps(render_kwargs, args_key):
    """Create and store maps for character data."""
    ### get available codes for the font used
    valid_codes = CHAR_CODE_MAP[render_kwargs["font_style"]]

    ### create surface map with commonly used chars if
    ### they are available in the font used; a width map
    ### with width of corresponding surfaces is also
    ### created;

    surf_map = {}
    width_map = {}

    for char in COMMON_CHARS:

        ## only consider char whose unicode value is within
        ## valid codes
        if ord(char) in valid_codes:

            ## then trying generating surf and filling the
            ## maps
            try:
                surf = safe_text_render(char, render_kwargs)

                width_map[char] = surf.get_width()
                surf_map[char] = surf

            ## if the surf wasn't generated, then the surf
            ## is None and will trigger an attribute error
            ## when trying to execute surf.get_width in the
            ## 'try' block, in which case the code for the
            ## char must be removed from the valid codes
            ## list
            except AttributeError:
                valid_codes.remove(ord(char))

    ### store maps and valid char codes list in the
    ### CHAR_DATA_MAP, referenced by the json configuration
    ### string
    CHAR_DATA_MAP[args_key] = (surf_map, width_map, valid_codes)


### class definition


class Label(BasicObject):
    """A text widget with automatic surface updating.

    It works similar to tkinter.Label objects since they
    can have text set and reset any time.
    """

    def __init__(
        self,
        text="",
        *,
        font_size=22,
        font_style="default",
        foreground_color=BLACK,
        background_color=(*BLACK, 0),
        coordinates_name="topleft",
        coordinates_value=(0, 0)
    ):
        """Perform setups and assign data for reuse.

        text
            string providing text to update surface.
        font_size
            Integer indicating desired font size.
        font_style
            String. Either 'default' for default font or
            'mono' for monospace font.
        foreground_color
        background_color
            A tuple or list of r, g, b values which are
            integers ranging from 0 to 255. The background
            color can be None, though, in which case the
            background is rendered transparent. For the
            background color, an optional fourth value can
            be passed, which is an integer in the same range
            as the others, representing the image opacity
            (0 for full transparency and 255 for full
            opacity).
        coordinates_name
            String representing attribute name of rect
            wherein to store the position information from
            the coordinates value parameter.
        coordinates_value
            Position information in the form of a tuple or
            list containing 2 integers representing
            positions in the x and y axes, respectively.
        """
        ### convert the colors passed into tuples for
        ### simplicity (since colors can appear in other
        ### classes like pygame.Color and builtin lists too)

        background_color = tuple(background_color)
        foreground_color = tuple(foreground_color)

        ### process the arguments so the data needed to
        ### manage the label is properly generated or
        ### retrieved

        ## the background is handled apart from the other
        ## arguments, so we just store it in its own
        ## attribute (since it is handled apart, notice we'll
        ## ommit it from the arguments in the next steps);
        self.bg_color = background_color

        ## the other arguments are gathered in a dict
        ## so they can be used as keyword arguments to
        ## render characters

        self.render_kwargs = {
            "font_size": font_size,
            "font_style": font_style,
            "foreground_color": foreground_color,
        }

        ## a tuple with the arguments is also created
        ## to be used as a key representing the specific
        ## combination of arguments; this way the data
        ## related to each different setting will never
        ## be generated more than once

        args_key = (font_size, font_style, foreground_color)

        ## the CHAR_DATA_MAP holds all the data needed to
        ## manage labels, for each different combination
        ## of arguments used;
        ## if said map lacks the data we need for our
        ## specific combination of arguments (the key we
        ## just put together), we create such data:
        if args_key not in CHAR_DATA_MAP:
            create_char_maps(self.render_kwargs, args_key)

        ## finally, we retrieve and assign the data we need
        ## from the CHAR_DATA_MAP using the key
        self.surf_map, self.width_map, self.valid_codes = CHAR_DATA_MAP[args_key]

        # TODO this step of storing the height shouldn't be
        # repeated every time an instance is created
        # (probably include it on the if block above)
        ### store the height of a random character
        self.height = choice(list(self.surf_map.values())).get_height()

        ### create and position rect

        self.rect = Rect((0, 0), (0, self.height))

        setattr(self.rect, coordinates_name, coordinates_value)

        ### also store the coordinate information

        self.coordinates_name = coordinates_name
        self.coordinates_value = coordinates_value

        ### create an attribute with a string to store the
        ### text of the label
        self.contents = ""

        ### add text the text passed as argument (this is also
        ### needed in order to create an image attribute and
        ### thus prevent errors)
        self.add_text(text)

        ### define update behaviour
        self.update = empty_function

    def get(self):
        """Return text content of widget."""
        return self.contents

    def set(self, text):
        """Set obj text.

        text (string)
            value to be used as label text.
        """
        ### return earlier if text is already set
        if text == self.contents:
            return

        ### otherwise "clear" contents and add text

        self.contents = ""
        self.add_text(text)

    def add_text(self, text):
        """Add text.

        text
            Any string obj.
        """
        ### only keep valid characters (present on map)
        filtered_text = "".join(list(filter(self.filter_char, text)))

        ### extend content with filtered text
        self.contents += filtered_text

        ### update image
        self.update_image()

    def filter_char(self, char):
        """Return True if a surf is available for char."""
        ### only proceed if code for char is valid
        ### (present on font)
        if ord(char) in self.valid_codes:

            ## if a surface was already generated for the
            ## char, them return True
            if char in self.surf_map:
                return True

            ## otherwise generate said surf and populate it
            ## the respective values in the surf map and
            ## width map
            else:

                # try generating surf and populating the
                # maps
                try:
                    surf = safe_text_render(char, self.render_kwargs)

                    self.width_map[char] = surf.get_width()
                    self.surf_map[char] = surf
                    return True

                # if the surf wasn't generated, then the
                # surf is None and will trigger an attribute
                # error when trying to execute
                # surf.get_width in the 'try' block, in
                # which case the code for the char must be
                # removed from the valid codes list
                except AttributeError:
                    self.valid_codes.remove(ord(char))

    def update_image(self):
        """Update image surface attribute."""
        ### create new surface for image attribute

        ## calculate width
        ## (if there are no contents, width might even be
        ## zero at this point)
        width = sum(map(self.width_map.get, self.contents))

        ## store dimensions with extra space for padding
        size = (width + 10, self.height + 10)

        ## generate and store surface
        self.image = render_rect(*size, self.bg_color)

        ### iterate over contents blitting surfaces

        ## define a starting x and y for blitting characters;
        ## notice they start with 5, to give the surface a
        ## small padding
        x = y = 5

        ## blit characters while incrementing the x position

        for char in self.contents:

            # blit surf
            self.image.blit(self.surf_map[char], (x, y))

            # increment x
            x += self.width_map[char]

        ### admin task: update rect dimensions and position

        self.rect.size = self.image.get_size()

        setattr(self.rect, self.coordinates_name, self.coordinates_value)



"""Facilities for Game class, the main application."""

### third party imports and setups

from pygame import init as init_pygame
from pygame.mixer import pre_init

## sound pre initialization
pre_init(44100, -16, 8, 1024)

## pygame initialization
init_pygame()

## extra pygame imports

from pygame import quit as quit_pygame, error as pygame_error

from pygame.event import clear as clear_events


### local imports and setups

from .config import CLOCK, FPS

from .music import MUSIC_MANAGER

### start music and blit a loading screen
MUSIC_MANAGER.update()

from .appcommon.exception import (
    QuitGameException,
    LevelSwitchException,
    RestartLevelException,
    ManagerSwitchException,
)
from .appcommon.behaviour.main import (
    get_level,
    load_game,
    get_custom_manager,
    reset_player_health,
)
from .appcommon.behaviour.buffer import remove_buffers

from .logconfig import APP_LOGGER

logger = APP_LOGGER.getChild(__name__)


# TODO the player probably only need to be instantiated
# once; should you do that? Also, the current level
# instance could be stored in the GAME_REFS? (in which
# case we should probably change the name of the object);
# probably instantiate player as you do with the
# music manager;

# TODO eliminate the concept of scripted level; just use
# game levels and the custom managers

# TODO since the event queue is populated even during
# level transitions (for instance, while loading assets
# when transitioning from logo screen to main menu),
# it should be a good idea to clear the event queue
# just as you begin looping throught a new level;
# could there be more objects like the fps display
# which only need to be instantiated once? think
# about it and take measures to guarantee you're not
# reinstantiating objects unnecessarily;

# TODO I must somehow prevent user from leaving screen
# after confirm saving prompt, or the game will freeze;

# XXX
# Before changing the level, the game.Game still
# finished the execution of the level manage_updates
# method, which cause a pygame.display.update to
# be executed and thus update the screen, redrawing
# what was there before. (Now that I think of it, this
# is somehow absurd, so I'll leave it here anyways but
# to describe the behaviour, not necessarily the cause,
# that is, I'm not sure of this myself,
# but this is what it seems to happen: before the
# next level starts we have a quick split second glimpse
# of screen last frame from last level).

# XXX
# task manager may very well be able to replace the current
# level switching timing feature and the restart level too
# (this last one is the one used whenever the player dies).
# Edit: this works but is a discarded idea, because since
# the task manager discards all tasks when switching
# between levels, the tasks are discarded while we are
# iterating over them.
# This must be properly archived for future reference.
# Since I don't understand completely the underlying
# workings of the "for i in iterable" block, I'm not sure
# if I can asses the situation accurately, but at least the
# current solution isn't bad at all, so it's ok. In fact,
# even if the tasks are discarded while iterating over them,
# it may be that this isn't a problem, since the iterable
# isn't changed itself, just 'unbound' from the task manager
# instance. Additionaly, the failed discarded attemp,
# even if it wasn't harmful, it was noticeable more slow,
# so the previous is the best anyway.

# XXX
# remember to eliminate the need to receive level references
# in classes that only used it to get screen references
# from its old screen attribute (not used anymore)

# XXX
# It would be great if I could change the default coordinates
# used to create objects (instead of just using
# bottomleft). See how you implement it.

# XXX
# The level switching could be done inside another
# very light update manager. This would eliminate the
# coexistence of two loaded levels at the same time
# whenever switching between them. Ponder.
#
# Edit: the temporary coexistence of two levels isn't
# as harmful as anticipated, because the surface data,
# which is the most memory consuming information, is always
# loaded in the memory. The only extra information kept
# simultaneously from both levels are regular 'object'
# subclasses and related data, which isn't memory consuming
# at all in comparison. Also, we never had any runtime
# problem related to memory, specially considering the
# machine in which the game is being developed is old and
# has only 2GB of memory, so we are very ok.

### main function


def run_game():
    """Run the game loop."""
    logger.info("Preparing to enter the game loop.")

    ### get logo screen scene
    update_manager = get_custom_manager("logo_screen.scn")

    ### create a flag to represent the game's running state
    running = True

    ### initialize the game loop
    while running:

        CLOCK.tick(FPS)

        ### try executing game loop specific methods of the
        ### update manager
        try:
            ### call the routine responsible for controlling
            ### the game
            update_manager.control()

            ### call the routines responsible for managing
            ### the game logic
            update_manager.update()
            MUSIC_MANAGER.update()

            ### call the routine responsible for drawing
            ### and updating the screen
            update_manager.draw()

        ### if either a level switch or manager switch
        ### exception is raised, assign the
        ### new_update_manager to the local variable;
        ### the exact type must be assessed though, so
        ### a proper message can be logged accordingly;

        except ManagerSwitchException as err:

            if type(err) is LevelSwitchException:
                logger.info("Switching between levels.")

            else:
                logger.info("Switching to update manager.")

            # TODO adjust code below

            manager_name = err.update_manager_name

            if isinstance(manager_name, str):
                if manager_name.endswith(".lvl"):
                    update_manager = get_level(manager_name)

                else:
                    update_manager = get_custom_manager(manager_name)

            else:
                update_manager = manager_name

            ### admin task: clear event queue
            clear_events()

        ### if a restart game exception is raised, execute
        ### the restart level function
        except RestartLevelException:

            remove_buffers()
            reset_player_health()

            update_manager = get_level(load_game())

        ### if a quit game exception is raised, trigger
        ### the exit of the game loop by assigning
        ### false to the running flag
        except QuitGameException:

            logger.info("Quitting game normally.")
            running = False

        ### general error handlinlg
        except (Exception, pygame_error) as err:
            print("here")
            logger.exception("Exitting game due to unexpected exception.")

            # XXX ponder about what to do here, there's
            # various possibilities and factors to consider
            # besides only calling quit_pygame (pygame.quit);

            quit_pygame()

            raise Exception from err

    ### remove swap files
    remove_buffers()

    ### log final call to pygame_quit and perform it

    logger.debug("pygame.quit being called.")
    quit_pygame()



"""Facility for sounds keeping."""

from os.path import join

### third-party import
from pygame.mixer import Sound

### local import
from .config import SOUNDS_DIR, USER_SETTINGS

### utility function


def sound_from_name(filename):
    """Create and return a Sound object.

    Locate it using its name.

    filename (string)
        represents the basename of the sound file.
    """
    return Sound(join(SOUNDS_DIR, filename))


### sounds map

SOUNDS_MAP = {
    "player_jump01": sound_from_name("player_jump_groan01.ogg"),
    "player_jump02": sound_from_name("player_jump_groan02.ogg"),
    "gui_step": sound_from_name("gui_step.ogg"),
    "game_start": sound_from_name("game_start.ogg"),
    "error": sound_from_name("error.ogg"),
    "key_typing01": sound_from_name("key_typing01.ogg"),
    "key_typing02": sound_from_name("key_typing02.ogg"),
}


def set_sounds_volume(volume):
    """Set the volume for all sounds in SOUNDS_MAP.

    volume (float)
        Float from 0.0 to 1.0.
    """
    for sound in SOUNDS_MAP.values():
        sound.set_volume(volume)


def get_sounds_volume():
    """Return the volume for any sound in SOUNDS_MAP."""
    SOUNDS_MAP[0].get_volume()


set_sounds_volume(USER_SETTINGS["sound_volume"])



"""Facility for the actors layer."""

from ...appcommon.task import add_task, add_sendoff_task
from ...appcommon.anim.player.main import AnimationPlayer

from .healthbar import HealthBar
from .thornbullet import ThornBullet


# XXX
# I still can't figure out how exactly json data is updated
# real time. I should schedule some experimentation so I
# can understand how everything works.
# edit: Now I understand it: I thought that by simply changing
# json data dict store as an attribute in the game objects
# the swap file were automatically updated. This isn't so.
# Changing the dict values changes the level_data dictionary
# which is an attribute of the level instance. What happens
# is that the level data is then saved before switching levels.
# The only change needed was making sure that the data in the
# json data attribute was changed before being saved in
# the swap file on disk.


class ThornShooter(object):
    """A thorn-shooting static plant."""

    level = None

    def __init__(self, prop_name, coordinates, health, revive_countdown):
        """Assign variables. and perform setups."""
        self.prop_name = prop_name

        self.anim_player = AnimationPlayer(
            self,
            prop_name,
            coordinates_name="bottomleft",
            coordinates_value=coordinates,
        )

        self.anim_player.play()

        self.health = health
        self.revive_countdown = revive_countdown

        self.set_operation_structure()

    def set_operation_structure(self):
        """Set other important operational structutes.

        Those include attributes and other setups.
        """

        self.healthbar = HealthBar(self)

        self.bullet = ThornBullet(self.rect.center, self.level)

        add_sendoff_task(self.store_data)
        self.pre_saving_routine = self.store_data

        if not self.health:
            self.revive_task = add_task(self.revive, self.revive_countdown)

    def update(self):
        """Update object state."""
        if self.health:
            self.bullet.update()

        self.anim_player.update()

    def draw(self):
        """Update object state."""
        if self.health:
            self.bullet.draw()

        self.anim_player.draw()
        self.healthbar.draw()

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)

    def die(self):
        """Perform death setups."""
        # XXX
        # some audio/visual feedback code should be here

        self.json_data["revive_countdown"] = self.revive_countdown = 10000
        self.revive_task = add_task(self.revive, self.json_data["revive_countdown"])

    def revive(self):
        """Revive the object."""
        # XXX
        # some audio/visual feedback
        # here
        self.json_data["health"] = self.health = 100
        self.json_data["revive_countdown"] = self.revive_countdown = 0
        self.healthbar.revive()
        self.bullet.reset_position()

    def store_data(self):
        """Store relevant json data."""
        self.json_data["health"] = self.health

        if not self.health:
            self.json_data["revive_countdown"] = self.revive_task.get_remaining()

        else:
            self.json_data["revive_countdown"] = 0

    def suffer_damage(self, damage):
        """Subtract damage from player healthbar."""
        # XXX
        # some audible and visual feedback is desired
        self.healthbar.update_width(-damage)



"""Facility for funcs to assist in anim data processing."""

### standard library

from types import SimpleNamespace
from operator import attrgetter
from itertools import permutations
from collections import defaultdict

### local imports

from ....common.tree import get_tree_values
from ....common.behaviour import get_nested_value


### quick utility
length_getter = attrgetter("length")

### functions


def get_filled_map(mapping, data, raw_data_key, post_process_key, func):
    """Return mapping after filling it with processed data.

    mapping (dict)
        mapping to hold the processed data.
    data (dict)
        mapping from which to retrieve the raw data for
        processing.
    raw_data_key, post_process_key (strings)
        they represent keys used to retrieve maps within the
        'data' argument. Such maps contain raw data to be
        processed and instructions to post-process the raw
        data processed from the first key respectively.
    func (callable)
        function to process the raw data.
    """
    for map_key in (raw_data_key, post_process_key):

        ### try retrieving raw data for processing
        try:
            retrieved_map = data[map_key]

        ### if absent, just break out of the for loop
        ### since the existence of both maps is
        ### optional and the absence of a map makes the
        ### presence of the next one unnecessary (if there's
        ### no data to process, then having post-process
        ### instructions has no meaning);
        except KeyError:
            break

        ### otherwise, process raw data
        else:
            for key, raw_data in retrieved_map.items():

                ## get processed data
                processed = func(raw_data, mapping)

                ## store the processed data in the mapping;

                ## note: in game applications we convert the
                ## processed data into a tuple before storing
                ## it; not here in the animation editor
                ## though, because if we did so editing live
                ## data would take much time due to
                ## numerous conversions from and to tuples;

                mapping[key] = processed

    return mapping


def manage_proxy_animations(
    metadata, structure, values, timing, surf_map, pos_map, timing_map
):
    """Setup proxy animations and related metadata.

    metadata (dict)
        stores animation metadata.
    structure (dict)
        contains structure information for each animation.
    values (dict)
        contains animation values for each animation and
        their respective nodes.
    timing (dict)
        contains timing information for each animation and
        their respective nodes.
    surf_map, pos_map, timing_map (dicts)
        contain animations value for surfaces, positions
        and animation timing, respectively.

    Proxy animations are animations produced by combining
    multiple pre-existent animation values and timing
    together. Such animation don't have animation data
    per se, except for structural data (which node is the
    parent/children of which).

    Thus, an animation is considered a proxy animation if
    there's no data for it in the values dict nor the
    timing dict.

    What we do here is:

        - Retrieve the proxy animations' names so we can
          reference them

        - If there are indeed proxy animations:
          - Reference the animation values and timing to
            which each node in the tree points (for the
            timing we actually generating a proxy-like/
            view-like object whose content is linked with
            the original walking deque but has independent
            walking; this is the ideal solutions, otherwise,
            if more than one node in the proxy animation
            were to reference the same timing deque, such
            deque would walk the animation value more than
            once per loop; for more information, visit the
            common/walkingdeque.py module);
    """
    ### retrieve the proxy animations' names
    proxy_names = metadata["proxy_anim_names"]

    ### store key names for different types of animation
    ### values and timing to use when building the proxy
    ### animations

    values_keys = ("surface_data", "position_data")
    indices_keys = ("surface_indices", "position_indices")

    ### store a map which associates raw values keys to
    ### their respective processed data ones

    processed_values_keys = ("surfaces", "positions")

    key_to_proc = dict(zip(values_keys, processed_values_keys))

    ### store map which associates values keys with their
    ### respective data map

    values_maps = (surf_map, pos_map)

    key_to_map = dict(zip(values_keys, values_maps))

    ### iterate over proxy animations (if there are any)
    ### referencing animation data from other animations
    ### for each node

    for anim_name in proxy_names:

        ## reference the tree structure and use it to
        ## retrieve the data for references for each kind
        ## of data

        tree = structure[anim_name]["tree"]

        refs_data = get_tree_values(tree, "reference_data", "children")

        ## iterate over each node and its respective
        ## referenced animation, creating references
        ## for animation values and timing

        # create mappings for proxy values and timing

        values[anim_name] = proxy_values = {}
        timing[anim_name] = proxy_timing = {}

        # reference node names of the proxy animation
        node_names = structure[anim_name]["node_names"]

        # finally iterate

        for node_name, ref_data in zip(node_names, refs_data):

            # create mappings for node anim values and timing

            proxy_values[node_name] = {}
            proxy_timing[node_name] = {}

            # assign value references

            for key in values_keys:

                # get data used to reference the values
                values_ref_data = ref_data[key]

                # get respective processed value key
                proc_key = key_to_proc[key]

                # if the reference data is a list, then it
                # must have length == 2; it contains the
                # names of animation and node referenced
                # respectively

                if isinstance(values_ref_data, list) and len(values_ref_data) == 2:

                    # unpack values into names of the
                    # referenced animation and node
                    ref_anim, ref_node_name = values_ref_data

                    # retrieve values using those names
                    retrieved_values = get_nested_value(
                        values, ref_anim, ref_node_name, proc_key
                    )

                # otherwise if the reference data is a
                # string, then it is a key to retrieve the
                # needed data in the respective value map

                elif isinstance(values_ref_data, str):

                    # retrieve specific value map for key
                    values_map = key_to_map[key]

                    # retrieve values using the reference
                    # data as key
                    retrieved_values = values_map[values_ref_data]

                # finally assign the values to the proxy
                # animation for the node in the proper key
                proxy_values[node_name][proc_key] = retrieved_values

            # assign timing "content content views"
            # (see common/wdeque/main.py for info on
            # content views)

            for key in indices_keys:

                # get data used to reference the timing
                timing_ref_data = ref_data[key]

                # if the reference data is a list, then it
                # must have length == 2; it contains the
                # names of animation and node referenced
                # respectively

                if isinstance(timing_ref_data, list) and len(timing_ref_data) == 2:

                    # unpack values into names of the
                    # referenced animation and node
                    ref_anim, ref_node_name = timing_ref_data

                    # retrieve timing view using those names
                    timing_view = get_nested_value(
                        timing, ref_anim, ref_node_name, key
                    ).get_content_view()

                # otherwise if the reference data is a
                # string, then it is a key to retrieve the
                # needed data in the timing map

                elif isinstance(timing_ref_data, str):

                    # retrieve walking deque from timing
                    # map
                    walking_deque = timing_map[timing_ref_data]

                    # retrieve the timing view
                    timing_view = walking_deque.get_content_view()

                # XXX add a function on sibling module
                # checks.py to ensure there's never a
                # scenario different from the if/elif blocks
                # above

                proxy_timing[node_name][key] = timing_view


def store_anim_clock_keys(metadata, timing):
    """Store keys to retrieve an animation clock.

    metadata
        Dictionary to store animation metadata.
    timing
        Dictionary containing timing information about
        animations for each node.

    For each animation, we measure the length of all
    timing sequences. The ones with the largest length
    for each animation have the node name and sequence
    name stored so that such sequence can be retrieved later.

    Such 'sequence' is in fact an instance of
    a collections.deque subclass defined on
    common/wdeque/main.py module.

    The largest one among them is called an animation clock,
    because we can use it to infer data like the
    current frame being played, how much frames it will
    take for the animation to complete the current loop
    and how many loops the animation performed (including
    backwards, if it is the case).

    Storing a reference to the deque itself would
    work too, but since all timing sequences are meant
    to be 'deepcopied' on the game package, we use the
    alternative approach of storing the keys instead.
    """
    metadata["largest_deques"] = largest_deques = {}

    length_zero_obj = SimpleNamespace(length=0)

    sequence_names = ["surface_indices", "position_indices"]

    for anim_name in timing:
        anim_timing_values = timing[anim_name]

        ### create map to store keys for the largest deque
        ### in the current animation
        largest_deques[anim_name] = anim_largest = {}

        ### Assign dummy as current largest deque
        largest_deque = length_zero_obj

        ### iterate over nodes' deques to find largest one
        for node_name in timing[anim_name]:
            node_sequences_map = anim_timing_values[node_name]

            ## group time deques together
            ## and retrieve largest

            deques = [largest_deque]

            for key in sequence_names:
                deques.append(node_sequences_map[key])

            largest_deque = max(deques, key=length_getter)

            ## If a deque from the node end up being
            ## the largest one, store the keys

            for key in sequence_names:
                if largest_deque == node_sequences_map[key]:
                    anim_largest["node_name"] = node_name
                    anim_largest["sequence_name"] = key


def store_root_pos_exchange_map(metadata, structure):
    """Generate/gather and store root position exchange map.

    metadata (dict)
        Storage for animation metadata.
    structure (dict)
        Contains structure information for each animation.

    A root position exchange map is a map which maps
    strings other dicts. Each string represents the name
    of a different root node in the animation (if the
    animations in the file have different root nodes).
    The dicts to which each string points contain the
    name of other root nodes pointing to a list of
    pygame.Rect attribute names.

    For instance:

    {
      "root_a": {
        "root_b": ["center", "center"]
      },
      "root_b": {
        "root_a": ["midbottom", "midbottom"]
      }
    }

    This map exists in order to prevent conflicts when
    exchanging positions between rects of different root
    nodes in the game/application space, specially when
    such nodes have rects of different sizes.
    """
    ### retrieve all root names and create pairs from the
    ### permutations between them

    root_names = {struct_data["tree"]["name"] for struct_data in structure.values()}

    pairs = permutations(root_names, 2)

    ### create the root pos exchange map using the pairs
    ### as nested keys and attributing the default value to
    ### each of them

    ## create a default dict
    root_pos_exchange_map = defaultdict(dict)

    ## iterate over pairs creating nested records of
    ## root position exchange data; such records are maps
    ## associating which position of root_a is used to
    ## which position of root_b; here position means the
    ## name of the pygame.Rect attribute from wherein to
    ## find/assign the position

    for root_a, root_b in pairs:
        root_pos_exchange_map[root_a][root_b] = ["center", "center"]

    ### update root pos exchange map values using values
    ### provided by the user, if any

    ## try retrieving user provided data
    try:
        root_pos_exchange_data = metadata["root_pos_exchange_map"]

    ## if not present, just pass
    except KeyError:
        pass

    ## otherwise, update values in the existing map
    else:
        root_pos_exchange_map.update(root_pos_exchange_data)

    ### finally store the exchange map in the metadata map
    metadata["root_pos_exchange_map"] = root_pos_exchange_map



"""Facility for prompt generation."""

from math import pi

### third-party imports

from pygame import Rect
from pygame.draw import (
    rect as draw_rect,
    circle as draw_circle,
    line as draw_line,
    arc as draw_arc,
    polygon as draw_polygon,
)

### local imports

from ..palette import BLACK, ARBITRARY_COLORKEY, PROMPT_BACKGROUND, has_same_rgb
from .surf import render_rect
from .text.main import render_text


def prompt_factory(
    prompt_text, object_pos, item_list, background_color=PROMPT_BACKGROUND
):
    """Return a list of interlocutor/surface pairs.

    ### ASCII art: generated surfaces take this final form:
    # obs.: + and * represent the area of different surfaces
    # that are drawn bellow the prompt text.
     _____________     __________________________
    (             )   (                          )
    | Want, this? |   | Should I take the items? |
    | +++         |   | +++  ***                 |
    | +++         |   | +++  ***                 |
    (_____________)   (__________________________)

    Notice the width and height are
    automatically adjusted to comprise the text surface
    and item surfaces.

    prompt_text
        A string representing the text of the prompt.

    object_pos
        Position of object generating the prompt.

    item_list
        List of surfaces representing the items offered
        in the prompt.

    background_color
        A tuple or list of r, g, b values which are
        integers ranging from 0 to 255. This color will
        be in several pygame.draw drawing calls.

        Though this function wasn't intended to be used
        in this way an optional alpha value within the
        same range can also be provided in the tuple
        or list, and the purpose of this can be read
        in the link bellow:
        https://www.pygame.org/docs/ref/draw.html

        Here's the specific excerpt:
        "Most of the arguments accept a color argument
        that is an RGB triplet. These can also accept
        an RGBA quadruplet. The alpha value will be
        written directly into the Surface if it contains
        pixel alphas, but the draw function will not draw
        transparently. The color argument can also be an
        integer pixel value that is already mapped to the
        Surface's pixel format."
    """
    if has_same_rgb(background_color, ARBITRARY_COLORKEY):
        msg = "'background_color' can't be equal to" + " 'ARBITRARY_COLORKEY' constant."
        raise ValueError(msg)

    bg_color = PROMPT_BACKGROUND
    common_width = 4

    ### Preparation to create initial surf to use as canvas

    line_str = prompt_text
    interlocutor_pos = object_pos

    text_surf = render_text(line_str, font_size=22, foreground_color=BLACK, padding=0)
    text_surf_height = text_surf.get_height()

    total_width = 0
    for item in item_list:
        total_width += item.get_width()

    box_width = (max(total_width, text_surf.get_width())) + 40

    # pygame don't handle lambdas well, so using this instead
    def get_height(item):
        """Return item height."""

    box_height = (
        max(item_list, key=get_height).get_height() + text_surf.get_height() + 20
    )

    surf = render_rect(box_width, box_height)
    surf.fill(ARBITRARY_COLORKEY)

    ### Prompt box volume (blocking)

    ## 02 rects
    # width and height depend on text and item surfaces
    # inside
    #
    #    rect1
    # |---------------|
    #     _________      _
    #  __|         |__   |
    # |__           __|  |  rect2
    #    |_________|     |
    #
    #
    hrect = Rect((0, 20), (box_width, box_height - 40))
    vrect = Rect((20, 0), (box_width - 40, box_height))
    draw_rect(surf, bg_color, hrect)
    draw_rect(surf, bg_color, vrect)

    ## 04 circles
    #
    # tlpos _________ trpos
    #    __.         .__
    #   |__.         .__|
    # blpos|_________| brpos
    #
    # each point above represents the center where
    # the circles will be drawn. This way it'll appear
    # that the box has round edges.
    #
    # Also, in next code blocks,
    # tl means topleft, tr means topright,
    # bl means bottomleft and  br means bottomright.
    tlpos, trpos, blpos, brpos = (
        (20, 20),
        (box_width - 20, 20),
        (20, box_height - 20),
        (box_width - 20, box_height - 20),
    )
    radius = 20
    draw_circle(surf, bg_color, tlpos, radius)
    draw_circle(surf, bg_color, trpos, radius)
    draw_circle(surf, bg_color, blpos, radius)
    draw_circle(surf, bg_color, brpos, radius)

    ### Prompt box outlines and final touches

    ## 04 arcs
    # will be drawn around the round corners of the
    # volume
    tl_arc_rect, bl_arc_rect, tr_arc_rect, br_arc_rect = (
        Rect((0, 0), (40, 40)),
        Rect((0, box_height - 40), (40, 40)),
        Rect((box_width - 41, 0), (40, 40)),
        Rect((box_width - 40, box_height - 40), (40, 40)),
    )

    startangle1, startangle2, startangle3, startangle4 = pi / 2, pi, 0, (3 * pi) / 2

    endangle1, endangle2, endangle3, endangle4 = pi, (3 * pi) / 2, pi / 2, 0

    draw_arc(surf, BLACK, tl_arc_rect, startangle1, endangle1, common_width)
    draw_arc(surf, BLACK, bl_arc_rect, startangle2, endangle2, common_width)
    draw_arc(surf, BLACK, tr_arc_rect, startangle3, endangle3, common_width)
    draw_arc(surf, BLACK, br_arc_rect, startangle4, endangle4, common_width)

    # eliminating artifacts in the drawn arcs
    for i in range(3):
        tl_arc_rect.y += 1
        draw_arc(surf, BLACK, tl_arc_rect, startangle1, endangle1, 1)
    for i in range(2):
        tl_arc_rect.x += 1
        draw_arc(surf, BLACK, tl_arc_rect, startangle1, endangle1, 1)

    for i in range(3):
        bl_arc_rect.y += -1
        draw_arc(surf, BLACK, bl_arc_rect, startangle2, endangle2, 1)
    for i in range(2):
        bl_arc_rect.x += 1
        draw_arc(surf, BLACK, bl_arc_rect, startangle2, endangle2, 1)

    for i in range(3):
        tr_arc_rect.y += 1
        draw_arc(surf, BLACK, tr_arc_rect, startangle3, endangle3, 1)
    for i in range(2):
        tr_arc_rect.x += -1
        draw_arc(surf, BLACK, tr_arc_rect, startangle3, endangle3, 1)
    tr_arc_rect.x += 3
    draw_arc(surf, BLACK, tr_arc_rect, startangle3, endangle3, 1)
    tr_arc_rect.y -= 3
    draw_arc(surf, BLACK, tr_arc_rect, startangle3, endangle3, 1)

    for i in range(3):
        br_arc_rect.y += -1
        draw_arc(surf, BLACK, br_arc_rect, startangle4, endangle4, 1)
    for i in range(2):
        br_arc_rect.x += -1
        draw_arc(surf, BLACK, br_arc_rect, startangle4, endangle4, 1)

    ## 04 lines
    # will be drawn to outline the straight sections
    # of the volume
    #
    # in the end, the box will be more or less like
    # this:
    #  _______________
    # (               )
    # |               |
    # (_______________)
    #
    start1, end1 = (20, 1), (box_width - 20, 1)
    start2, end2 = ((19, box_height - 3), (box_width - 20, box_height - 3))
    start3, end3 = (1, 19), (1, box_height - 22)
    start4, end4 = ((box_width - 3, 20), (box_width - 3, box_height - 21))
    draw_line(surf, BLACK, start1, end1, common_width)
    draw_line(surf, BLACK, start2, end2, common_width)
    draw_line(surf, BLACK, start3, end3, common_width)
    draw_line(surf, BLACK, start4, end4, common_width)

    ### Blitting the text and making other adjustments

    surf.blit(text_surf, (19, 8))

    # blitting the item surfaces
    x = 19
    for item in item_list:
        surf.blit(item, (x, text_surf_height + 10))
        x += item.get_width()

    # Making everything else outside box volume transparent
    surf.set_colorkey(ARBITRARY_COLORKEY)

    prompt_pair = (interlocutor_pos, surf)

    return prompt_pair



"""Facility for json related tools."""

from json import load, loads, dump


def load_json(filepath):
    """Return json data from file in filepath.

    filepath
        String representing absolute or relative path of
        json file to be loaded.
    """
    try:
        with open(filepath, mode="r", encoding="utf-8") as f:
            data = load(f)
    except FileNotFoundError as err:
        print("File not found: {}".format(filepath))
        raise err

    return data


def save_json(data, filepath):
    """Save data in filepath as a json file.

    data
        Any json serializable data.
    filepath
        String representing absolute or relative path of
        json file to be loaded.
    """
    has_backup = False

    ### before saving data, retrieve a backup of the file
    ### if it exists
    try:
        with open(filepath, mode="r", encoding="utf-8") as f:
            backup = f.read()
            has_backup = True

    except FileNotFoundError as err:
        if not filepath in str(err):
            raise err

    ### only then try to save
    try:
        with open(filepath, mode="w", encoding="utf-8") as w:
            dump(data, w, ensure_ascii=False, indent=2, sort_keys=True)

    ### if something go wrong, save the backup back
    except Exception as err:

        if has_backup:
            with open(filepath, mode="w", encoding="utf-8") as w:
                w.write(backup)

        raise Exception from err



"""Facility for node class."""

from pygame import Rect

from ...common.math import invert_point
from ..autoblit import BasicObject


class Node(BasicObject):
    """Provides operations for tree-structured animation.

    This class is used to represent an object as part of a
    compound object, that is a node in a tree of objects.

    It provides operations to change its positions over time
    relative to a parent node which can be (re)assigned
    anytime to the 'parent' attribute.

    Thanks to that, it can be easily (re)arranged into
    multiple combinations with other instances into a
    tree structure just by having a parent assigned to each
    node, except for the 'top' parent (the root), which
    should have 'None' assigned to the parent attribute in
    order to work properly. This is how this object is
    intended to be used.
    """

    def __init__(self, node_name, node_data):
        """Store data and perform setups.

        ### Arguments

        node_name (string)
            Unique name of the node.

        node_data (dict)
            data about the node's dimensions and, if it has
            collision data, the dimensions of the collision
            rect and its offset relative to the center of
            the node rect, if any.
        """
        ### store unique name
        self.name = node_name

        ### process geometry related data

        geometry_data = node_data["geometry"]

        ## Create normal rect/image attributes
        width, height = geometry_data["dimensions"]
        self.rect = Rect((0, 0), (width, height))

        ## Process collision data if existent

        try:
            collision_data = geometry_data["collision_data"]

        except KeyError:
            self.col_rect = None

        else:
            ## Create col_rect attribute for collisions

            width, height = collision_data["dimensions"]
            self.col_rect = Rect(self.rect.topleft, (width, height))
            self.col_rect_offset = collision_data.get("offset", (0, 0))

    def set_pos(self, relative_pos):
        """Set node position relative to parent.

        relative_pos (2-tuple of integers or list with 2 integers)
            Contains x and y distance of this node relative
            to the parent, center to center.
        """
        ### try moving the rect center to the moved center
        ### of the parent
        try:
            self.rect.center = self.parent.rect.move(relative_pos).center

        ### if parent is None (an attribute error will be
        ### raised trying to access 'rect') check if rect
        ### needs to be offset relative to col_rect
        except AttributeError:

            ## try offseting rect relative to collision rect

            try:

                # invert collision rect offset
                inverted_pos = invert_point(self.col_rect_offset, True, True)

                # offset rect
                self.rect.center = self.col_rect.move(inverted_pos).center

            ## do nothing if inexistence of col_rect_offset
            ## attribute causes an error
            except AttributeError:
                pass

        ### if everything went ok, we just
        ### check if col_rect needs to be positioned
        else:

            ## try positioning col_rect relative to rect.

            try:
                self.col_rect.center = self.rect.move(self.col_rect_offset).center

            ## do nothing if inexistence of col_rect_offset
            ## attribute causes and error
            except AttributeError:
                pass

    def get_distance_from_parent(self):
        """Return 2-tuple of x and y distance from parent.

        The distance is measured between the center of the
        node's rect and parent's rect.
        """
        ### get the inverted center of parent

        inv_parent_center = invert_point(self.parent.rect.center, True, True)

        ### return the center of a rect obtained from
        ### moving the node rect towards the inverted
        ### coordinates of the parent's rect

        return self.rect.move(inv_parent_center).center



"""HealthBar class facility for player health management."""

### third-party imports

from pygame import Rect
from pygame.draw import rect as draw_rect

### local imports

from ..config import SCREEN
from ..palette import BLACK
from ..common.behaviour import empty_function
from ..appcommon.surf import render_rect


class HealthBar:
    """A health display bar."""

    def __init__(self, player):
        """Initialize superclass and set variables.

        player
            A reference to player.main.Player instance.
        """
        self.player = player

        self.image = render_rect(200, 20)
        self.image.fill((255, 90, 40))

        self.rect = self.image.get_rect()
        self.topleft = (160, 60)

        outline_x, outline_y = (self.topleft[0] - 2, self.topleft[1] - 2)
        self.outline_rect = Rect((outline_x, outline_y), (204, 24))

        self.max_health = 100
        self.min_health = 0

        self.is_dead = False

        self.update_width()

        self.update = empty_function

    def update_width(self, amount=0):
        """Increase or decrease health by amount.

        It also updates size according to current health.

        amount (defaults to 0)
            An integer expressing the amount to add or
            subtract (if it is negative) from health.
        """
        self.player.health += amount
        self.player.health = max(
            min(self.player.health, self.max_health), self.min_health
        )
        self.rect.width = self.player.health * 2

        if not self.player.health and not self.is_dead:
            self.player.die()
            self.is_dead = True

    def draw(self):
        """Update object state."""
        # XXX
        # It is better to later change this to use a
        # mechanism similar to thornshooter healthbar,
        # (that is, the stencil-like feature)
        # since this current area rect feature was
        # implemented before I fully understood it. I don't
        # find it elegant/simple anymore.
        SCREEN.blit(self.image, self.topleft, self.rect)
        draw_rect(SCREEN, BLACK, self.outline_rect, 4)



"""Facility for the back_props layer."""

from os.path import join, dirname

### third-party imports

from pygame.draw import rect as draw_rect
from pygame.image import save as save_image
from pygame.transform import scale as scale_surface

### local imports

from ..config import SCREEN

from ..appcommon.surf import INVISIBLE_SURF
from ..appcommon.task import add_task
from ..appcommon.overrider import define_script
from ..appcommon.promptfactory import prompt_factory
from ..appcommon.behaviour.save import save_game
from ..appcommon.anim.player.main import AnimationPlayer

from ..common.math import unscroll_coordinates

from ..palette import BLACK


# XXX
# This class animation uses repeated unnecessarily
# pre-created surfaces (like transparent ones),
# repeating them too. Fix this.


class SatelliteDish:
    """A satellite dish which performs progress saving."""

    level = None

    def __init__(self, prop_name, coordinates):
        """Initialize superclass and variables.

        prop_name
            A string representing the prop name.
        coordinates
            List or tuple of x and y coordinates representing
            the desired bottomleft coordinates for the
            rect attribute.
        """
        self.prop_name = prop_name

        self.anim_player = AnimationPlayer(
            self,
            prop_name,
            coordinates_name="bottomleft",
            coordinates_value=coordinates,
        )

        self.anim_player.play()

        self.set_operational_structure()

    def set_operational_structure(self):
        """Set additional objects with which to work."""
        ## prepare prompt
        object_screen_midtop = unscroll_coordinates(self.rect.midtop)

        offset_screen_midtop = (object_screen_midtop[0], object_screen_midtop[1] - 32)
        item_list = [INVISIBLE_SURF]
        prompt_pair = prompt_factory(
            "Should I save my progress?", offset_screen_midtop, item_list
        )
        prompt_action = self.trigger_game_saving
        self.prompt = (prompt_pair, prompt_action)

    def update(self):
        """Update object state."""
        self.manage_animations()
        self.anim_player.update()

    def draw(self):
        """Draw animation."""
        self.anim_player.draw()

    def manage_animations(self):
        """Select suitable animation according to state."""
        anim_name = self.anim_player.anim_name
        loops_no = self.anim_player.peek_next_loops_no()

        if anim_name == "activating":

            if loops_no >= 1:
                self.anim_player.switch_animation("activated")

        elif anim_name == "activated":

            if loops_no >= 3:

                self.save_game()

                # XXX
                # Before this must play deactivating anim
                # maybe should also make player say
                # something like "great, now my progress is
                # saved!"
                # Also add sounds etc. in the method where
                # needed.
                self.anim_player.switch_animation("idle")
                self.level.get_control()

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)

    def receive_interaction(self, player):
        """Send prompt to player state manager."""
        player.send_prompt(self.prompt)

    def trigger_game_saving(self):
        """Setup a game saving script and execute it."""
        define_script("stand_still")

        self.level.give_up_control()
        self.anim_player.switch_animation("activating")

    def save_game(self):
        """Save game and level states and give feedback."""
        self.level.pre_saving_routine()
        self.level.save_states_buffer()

        save_game()

        add_task(self.store_thumbnail, 0, unit="frames")

        print("The game was saved.")

    def store_thumbnail(self):
        """Stores a thumbnail in the save directory."""
        save_dir = dirname(self.level.save_state_filepath)
        thumb_path = join(save_dir, "thumb.png")

        thumb = scale_surface(SCREEN, (640, 360))
        draw_rect(thumb, BLACK, thumb.get_rect(), 5)
        save_image(thumb, thumb_path)



"""Facility for the back_props layer."""

### local imports

from ..common.behaviour import empty_function
from ..appcommon.surf import render_image
from ..appcommon.autoblit import BasicObject


class StaticBackProp(BasicObject):
    """A simple static back prop."""

    prop_map = {
        "dirt_back_prop": render_image("dirt_block01.png"),
        "dirt_darker_back_prop": render_image("dirt_darker_back_prop.png"),
        "stone_darker_back_prop": render_image("stone_darker_back_prop.png"),
        "sand_back_prop": render_image("sand_back_prop.png"),
        "stone_back_prop": render_image("stone_back_prop.png"),
    }

    def __init__(self, prop_name, coordinates):
        """Assign variables.

        prop_name
            string representing the name of the prop.
        coordinates
            bottomleft coordinates for the rect attribute.
        """
        self.prop_name = prop_name

        self.image = StaticBackProp.prop_map[prop_name]

        self.rect = self.image.get_rect()

        self.rect.bottomleft = coordinates

        self.update = empty_function

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Facility with custom objects for easy blitting handling."""

from ..common.wdeque.main import WalkingDeque

### local import
from ..config import SCREEN, SCREEN_RECT


### callables to be bounded on custom containers


def update_objs(self):
    """Call the update method on all objects."""
    for obj in self:
        obj.update()


def draw_objs(self):
    """Call the draw method on all objects."""
    for obj in self:
        obj.draw()


def draw(self):
    """Blit obj.image at obj.rect coordinates.

    Naturally, it should only be used if objects have
    both an image attribute containing a pygame.Surface
    instance and a rect attribute containing a
    pygame.Rect instance.
    """
    for obj in self:
        SCREEN.blit(obj.image, obj.rect)


def draw_on_screen(self):
    """Works as self.draw, but check if obj is onscreen.

    Naturally, it should only be used if objects have
    both an image attribute containing a pygame.Surface
    instance and a rect attribute containing a
    pygame.Rect instance.
    """
    for obj in self.get_on_screen():
        SCREEN.blit(obj.image, obj.rect)


def draw_on_surf(self, surf):
    """Work as self.draw, but blit on specified surf.

    surf
        Any pygame.Surface instance. If such instance is
        also the surface returned by pygame.display.set_mode
        then it is recommended to use the self.draw
        or self.draw_on_screen method instead.
    """
    for obj in self:
        surf.blit(obj.image, obj.rect)


def draw_contained(self, rect):
    """Draw obj on screen if contained in rect.

    rect
        Any instance of pygame.Rect.
    """
    for obj in self:
        if rect.contains(obj.rect):
            SCREEN.blit(obj.image, obj.rect)


def get_on_screen(self):
    """Return iterator with all objects on screen.

    Objects on screen are those which collide with it.
    """
    return (obj for obj in self if SCREEN_RECT.colliderect(obj.rect))


def get_colliding(self, rect):
    """Return iterator of objects which collide with rect.

    rect (instance of pygame.Rect)
    """
    return (obj for obj in self if rect.colliderect(obj.rect))


def get_contained(self, rect):
    """Return iterator of objects contained in the rect.

    rect (instance of pygame.Rect)
    """
    return (obj for obj in self if rect.contains(obj.rect))


def get_unionrect(self):
    """Return union rect of all objects' rects."""
    if len(self):
        rects = [obj.rect for obj in self]
        return rects[0].unionall(rects[1:])

    else:
        raise RuntimeError("Container is empty.")


def get_top(self):
    """Return smaller top coordinate from objects rects."""
    return min(obj.rect.top for obj in self)


def get_bottom(self):
    """Return larger bottom coordinate from objects rects."""
    return max(obj.rect.bottom for obj in self)


def get_left(self):
    """Return smaller left coordinate from objects rects."""
    return min(obj.rect.left for obj in self)


def get_right(self):
    """Return larger right coordinate from objects rects."""
    return max(obj.rect.right for obj in self)


def get_width(self):
    """Return collective width."""
    return self.get_right() - self.get_left()


def get_height(self):
    """Return collective height."""
    return self.get_bottom() - self.get_top()


def move_rects(self, dx, dy):
    """Move all objects' rects in place by dx and dy.

    dx, dy
        Integers indicating how much in pixels the rect of
        each object must be moved."""
    for obj in self:
        obj.rect.move_ip(dx, dy)


def position_rects(self, attr_name, point):
    ### get the union rect
    union_rect = self.get_unionrect()

    ### copy and position union rect
    pos_union_rect = union_rect.copy()
    setattr(pos_union_rect, attr_name, point)

    ### move group objects so topleft points are the same

    start_x, start_y = union_rect.topleft
    end_x, end_y = pos_union_rect.topleft

    delta_x = end_x - start_x
    delta_y = end_y - start_y

    self.move_rects(delta_x, delta_y)


def scroll(self, x, y):
    """Call the scroll method on all objects."""
    for obj in self:
        obj.scroll(x, y)


def collide(self, to_collide):
    """Return all objects colliding with to_collide."""
    # XXX
    # This method will likely change in the near
    # future to accomodate different kinds of objects
    # by calling their own collision methods since
    # objects will now be much more dynamic/different in
    # their nature, assuming various configurations.
    colliding_objs = BlitterSet()
    for obj in self:
        if to_collide.rect.colliderect(obj.rect):
            colliding_objs.add(obj)

    return colliding_objs


def collide_any(self, to_collide):
    """Return an object of self if collides with obj."""
    # XXX
    # This method will likely change in the near
    # future to accomodate different kinds of objects
    # by calling their own collision methods since
    # objects will now be much more dynamic/different in
    # their nature, assuming various configurations.

    for obj in self:
        if to_collide.rect.colliderect(obj.rect):
            return obj

    return None


# XXX
# maybe make more functions to retrieve more position
# data from the object's rects inside containers;
# just using the union rect would be easier, but I wonder
# if it isn't quicker in some cases to just calculate the
# data you need. Must try and profile the options whenever
# convenient;

### custom containers


class BlitterSet(set):
    """Set with methods to draw/update objects."""


class BlitterList(list):
    """List with methods to draw/update objects."""


class BlitterWalkingDeque(WalkingDeque):
    """WalkingDeque with methods to draw/update objects."""


for cls in (BlitterSet, BlitterList, BlitterWalkingDeque):
    cls.update_objs = update_objs
    cls.draw = draw
    cls.draw_objs = draw_objs
    cls.draw_on_screen = draw_on_screen
    cls.draw_on_surf = draw_on_surf
    cls.draw_contained = draw_contained
    cls.get_on_screen = get_on_screen
    cls.get_colliding = get_colliding
    cls.get_contained = get_contained
    cls.get_unionrect = get_unionrect
    cls.get_left = get_left
    cls.get_right = get_right
    cls.get_width = get_width
    cls.get_top = get_top
    cls.get_bottom = get_bottom
    cls.get_height = get_height
    cls.move_rects = move_rects
    cls.position_rects = position_rects
    cls.scroll = scroll
    cls.collide = collide
    cls.collide_any = collide_any

### basic object


class BasicObject:
    """Basic object with methods for blitting."""

    def draw(self):
        """Blit itself on screen w/ image and rect attributes.

        Naturally, it should only be used if objects have
        both an image attribute containing a pygame.Surface
        instance and a rect attribute containing a
        pygame.Rect instance.
        """
        SCREEN.blit(self.image, self.rect)

    def draw_on_screen(self):
        """Works as self.draw, but check if obj is onscreen.

        Naturally, it should only be used if objects have
        both an image attribute containing a pygame.Surface
        instance and a rect attribute containing a
        pygame.Rect instance.
        """
        if SCREEN_RECT.colliderect(self.rect):
            SCREEN.blit(self.image, self.rect)

    def draw_on_surf(self, surf):
        """Work as self.draw, but blit on specified surf.

        surf
            Any pygame.Surface instance. If such instance is
            also the surface returned by
            pygame.display.set_mode then it is recommended
            to use the self.draw or self.draw_on_screen
            method instead.
        """
        surf.blit(self.image, self.rect)



"""Task processment facility for the scene player."""

from functools import partial

### local import
from ..appcommon.task import add_task


# TODO finish this


def process_tasks(obj, tasks):
    """Add tasks to the update manager related to the object.

    obj (either a sceneplayer.simple.SimpleObject or a
         sceneplayer.animated.AnimatedObject instance)

    tasks (list)
        each item contain data to create a task.
    """
    for data in tasks:

        ### create executable according to task name

        name = data["name"]

        if name == "visibility_on":
            callable_ = obj.visibility_on

        elif name == "visibility_off":
            callable_ = obj.visibility_off

        elif name == "switch_anim":

            anim_name = data["anim_name"]

            callable_ = partial(obj.anim_player.switch_animation, anim_name)

        ### add task with proper time unit

        try:
            delta_t = data["milliseconds"]
            unit = "milliseconds"

        except KeyError:
            delta_t = data["frames"]
            unit = "frames"

        add_task(callable_, delta_t, unit)



"""Log config for main application and its modules.

This module is maintained in another project. Future
improvements might include the generation of actual json
files, since as it is now it only generated json objects
and append them to the log file (instead of appending
them to an array and writing the array back as it would
be natural in a json file). Better control of naming would
also be great.

Additionally, large/verbose sentences, though useful are
difficult to present in various different medias. It would
be great if we could split a very large log into smaller
parts and log them all (at least in theory, this should
be tested in practice).
"""

from pathlib import Path

from os import environ

from os.path import join
from datetime import datetime
from json import dumps
from traceback import format_exception
from logging import Formatter, getLogger, DEBUG
from logging.handlers import RotatingFileHandler
from logging.handlers import TimedRotatingFileHandler


if "APPDATA" in environ:
    general_log_dir = Path(environ["APPDATA"])

else:
    general_log_dir = Path(environ["HOME"]) / ".local"

APP_LOGS_DIR = general_log_dir / "danittr" / "logs"

if not APP_LOGS_DIR.exists():
    APP_LOGS_DIR.mkdir(parents=True)

### Log level setup
# Change this value between the default levels
# depending on need. Use INFO level on normal operation.
LOG_LEVEL = DEBUG

APP_LOGGER = getLogger("app")
APP_LOGGER.setLevel(LOG_LEVEL)


### Custom formatter and tools


def json_serialize_traceback(exc_info):
    """Return json serializable version of exc_info"""
    exc_type, exc_value, exc_traceback = exc_info
    string_list = format_exception(exc_type, exc_value, exc_traceback)
    concatenated_string = "".join(string_list)
    return concatenated_string


class JsonLogFormatter(Formatter):
    """Outputs json logs."""

    def format(self, record):
        """Return a log record as json formated data.

        record
            A log record object automatically created."""
        if record.exc_info != None:
            exc_info = json_serialize_traceback(record.exc_info)
        else:
            exc_info = record.exc_info

        message = dict(
            name=record.name,
            timestamp=str(datetime.utcnow()),
            level=record.levelname,
            message=record.msg,
            exc_info=exc_info,
            lineno=record.lineno,
        )
        return dumps(message, indent=2, ensure_ascii=False)


json_formatter = JsonLogFormatter()


### Handlers


def custom_namer(name):
    """Custom namer function for handlers.

    name
        string automatically generated by the handlers.
    """
    return name + ".log"


## Handler to generate a record for each run

latest_filepath = str(APP_LOGS_DIR / "latest_run.log")
latest_run_handler = RotatingFileHandler(
    filename=latest_filepath, mode="a", maxBytes=0, backupCount=10, encoding="utf-8"
)

## Handler to generate a record for each day,
## regardless of quantity of runs
dailylog_path = str(APP_LOGS_DIR / "daily_runs.log")
daily_run_handler = TimedRotatingFileHandler(
    filename=dailylog_path,
    when="midnight",
    interval=1,
    backupCount=16,
    encoding="utf-8",
    utc=True,
)

### Final setup

latest_run_handler.setLevel(LOG_LEVEL)
latest_run_handler.setFormatter(json_formatter)
latest_run_handler.namer = custom_namer

daily_run_handler.setLevel(LOG_LEVEL)
daily_run_handler.setFormatter(json_formatter)
daily_run_handler.namer = custom_namer


APP_LOGGER.addHandler(latest_run_handler)
APP_LOGGER.addHandler(daily_run_handler)

APP_LOGGER.handlers[0].doRollover()

logger = APP_LOGGER.getChild(__name__)
logger.info("Configured logging.")



"""Facility for custom deque class definition."""

from collections import deque
from functools import partialmethod
from copy import deepcopy

# XXX
# replace the walking deque implementations in other
# packages by this one, since it is much more advanced;

### main class definition


class WalkingDeque(deque):
    """A deque with custom rotation and memory operations.

    The complete docstring with plenty doctests can be
    found on common/wdeque/wdtests.py
    """

    def __init__(self, iterable):
        """Initialize superclass, perform additional setups.

        Extends collections.deque.__init__.

        iterable (any non-empty iterable object)
           Contains items for the walking deque.
        """
        ### initialize the superclass; notice a maxlen
        ### argument isn't provided, we don't use it
        ### (check the class docstring for details)
        super().__init__(iterable)

        ### store length
        self.length = len(self)

        ### raise value error if iterable is empty
        if not self.length:
            raise ValueError("iterable must not be empty")

        ### create attributes to support operations; they
        ### will work as accumulators regarding both the
        ### amount they represent and the direction (their
        ### signal);

        self.total_walking = 0
        self.loops_no = 0

    def walk(self, steps):
        """Perform custom rotation operation.

        steps (integer)
            number of steps to 'walk' the deque.

        Extends collections.deque.rotate by inverting its
        direction and performing additional setups.
        The class docstring has complete explanation and
        examples (doctests).
        """
        ### perform superclass rotation using steps with
        ### inverted signal
        super().rotate(-steps)

        ### accumulate the steps in the total walking
        ### attribute and also use it to store how many
        ### times we walked through all the items

        self.total_walking += steps
        self.loops_no = self.total_walking / self.length

    def restore_walking(self):
        """Walk the deque back to its initial rotation."""
        self.walk(-self.total_walking)

    def get_mem_value(self, original_index):
        """Return value from index as if had never walked.

        original_index (integer)
            original index of the value you want to retrieve.
            This is the index of any value in the deque as
            if the deque had never walked.

        Same as __getitem__ if no rotation had ever happened.

        ### [Off-topic] Interesting behaviour

        An interesting behaviour occurs if we replace the
        line:

        value = self[original_index]

        by:

        self.walk(original_index)
        value = self[0]
        self.walk(-original_index)

        What will happen is that, if the index passed is
        out of range, it will return a value as if the
        deque kept rotating to compensate the missing
        indices.

        For instance:

        wd = WalkingDeque([0, 1, 2]),

        wd.get_mem_value(-4) would return 2;
        wd.get_mem_value(-5) would return 1;
        wd.get_mem_value( 3) would return 0;
        wd.get_mem_value( 4) would return 1;
        and so on...
        """
        ### store total walking
        tw = self.total_walking

        ### walk back the total walking to assume the
        ### original rotation
        self.walk(-tw)

        ### now that the deque is in its original walking,
        ### store the value in the provided index, which is
        ### the value we desire
        value = self[original_index]

        ## now, before returning the value, revert back to
        ## the previous rotation
        self.walk(tw)

        ### return the desired value
        return value

    def get_original_index(self, index):
        """Return original index of value in given index.

        The original index is the index of the value stored
        in the given index as if the deque never walked.

        index (integer)
            current index of the value whose original index
            you want to retrieve.
        """
        ### check if index is within range; this is quicker
        ### than "-self.length <= index < self.length";
        ### always remember efficiency is our priority;
        self[index]

        ### the original index is equivalent to the
        ### remainder of the sum of the current index and
        ### the total walking divided by the length
        return (index + self.total_walking) % self.length

    get_index_of_first = partialmethod(get_original_index, 0)

    def get_unwalked_list(self):
        """Return an unwalked list of the contents.

        Useful for collecting the original sorting
        of the elements without having to manually
        walk back and forth.
        """
        ### store total walking
        tw = self.total_walking

        ### between walking forth and back to the original
        ### rotation, store an unrotated list

        self.walk(-tw)
        unwalked_list = list(self)
        self.walk(tw)

        ### then return it
        return unwalked_list

    def peek_loops_no(self, steps):
        """Return loops number after temporary rotation.

        Used, for instance, to check if we'll complete a
        specific number of loops after a given number
        of rotations.

        steps (integer)
            number of steps back or forth to walk before
            peeking.

        We walk a number of times, store the number of
        loops performed up to that walking, then revert
        back to original walking, then return the stored
        number of loops.
        """
        ### rotate a given number of steps
        self.walk(steps)

        ### store the number of loops in this rotation
        peeked_loops_no = self.loops_no

        ### revert the number of steps performed
        self.walk(-steps)

        ### finally return the peeked loops_no
        return peeked_loops_no

    def get_content_view(self):
        """Return a special copy of the deque.

        Such copy has its walking state totally independent
        from the original object, but it is automatically
        updated every time the original has an item
        reassigned. For the update, the current index is
        not considered, but rather the original index, the
        one before every walk operation, which means their
        original rotation are always equal.
        """
        ### instantiate content view
        content_view = ContentView(self)

        ### append the content view to the list in the views
        ### attribute
        try:
            self.views.append(content_view)

        ### if such attribute doesn't exist, just store a new
        ### list in it and append the content view
        except AttributeError:
            self.views = []
            self.views.append(content_view)

        ### finally return the content view
        return content_view

    def __setitem__(self, index, value):
        """Set new value and update views, if any.

        Extends collections.deque.__setitem__

        index (integer)
            the existing index number of the item in the
            deque whose value you want to replace.

        value (any python value/object)
            object/value to replace the current object/value
            stored in the index.

        The index update in the views is the original index,
        not the current one.
        """
        super().__setitem__(index, value)

        ### try retrieving the views list
        try:
            views = self.views

        ### if no view exists, just pass
        except AttributeError:
            pass

        ### otherwise iterate views, updating them
        else:
            orig_index = self.get_original_index(index)
            for view in self.views:
                view.set_mem_value(orig_index, value)

    ### custom deepcopy operation

    def __deepcopy__(self, memo):
        """Return a deepcopy of this walking deque.

        This method is executed automatically by the
        copy.deepcopy function.

        memo (dict)
            provided automatically by the copy.deepcopy
            function. Used by copy.deepcopy to keep track
            of objects which were already deepcopied, thus
            avoiding deepcopying the same object more than
            once.
        """
        ### gather objects/data needed

        cls = self.__class__
        unwalked_items = self.get_unwalked_list()

        ### instantiate new object
        new_walking_deque = cls(unwalked_items)

        ### record (id(self) -> new_walking_deque) data pair
        ### in the memo mapping
        memo[id(self)] = new_walking_deque

        ### try deepcopying content views
        try:
            ## deepcopy each view;
            ## notice we don't need to catch a reference
            ## to the deepcopied views, since they are
            ## automatically stored in the views attribute
            for view in self.views:
                deepcopy(view, memo)

        ### if an attribute error is raised, it means we
        ### have no content views, so we just pass
        except AttributeError:
            pass

        return new_walking_deque

    ### custom string representation

    def __repr__(self):
        """Return an unambiguous string representation.

        Extends collections.deque.__repr__.
        """
        return super().__repr__().replace("deque", "WalkingDeque", 1)

    ### behaviour for suppressed methods

    def _suppressed(self, *args, **kwargs):
        """Behaviour for suppressed operations (methods)."""
        raise RuntimeError("operation ruled out by the design")

    ## suppress all methods ruled out by the class design
    ## by overriding them with the _suppressed method

    append = (
        appendleft
    ) = (
        extend
    ) = (
        extendleft
    ) = (
        pop
    ) = (
        popleft
    ) = (
        rotate
    ) = insert = clear = remove = reverse = __delitem__ = __delattr__ = _suppressed


### subclass definition


class ContentView(WalkingDeque):
    """View of walking deque content w/ autonomous walking.

    The complete docstring with plenty doctests can be
    found on common/wdeque/cvtests.py
    """

    def __init__(self, walking_deque):
        """Initialize super class and store proxied deque.

        Extends WalkingDeque.__init__.

        walking_deque
        (instance of common.walkingdeque.WalkingDeque)
            walking deque to be proxied.
        """
        super().__init__(walking_deque.get_unwalked_list())
        self.proxied = walking_deque

    def set_mem_value(self, original_index, value):
        """Set value for index as if we never "walked".

        original_index (integer)
            the existing original index of the item in the
            deque whose value you want to replace.

        value (any python value/object)
            object/value to replace the current object/value
            stored in the original index.
        """
        ### store total walking
        tw = self.total_walking

        ### walk back the total walking to assume the
        ### original rotation
        self.walk(-tw)

        ### now that the deque is in its original position,
        ### set the value in the provided index, which has
        ### the value we want to replace)
        self[original_index] = value

        ## now, revert back to teh previous rotation
        self.walk(tw)

    def __setitem__(self, index, value):
        """Set item value only if it complies to design.

        Extends WalkingDeque.__setitem__.

        According to the design, the value can only be set
        if it is identical (a is b) to the corresponding
        original index value of the proxied deque.

        index (integer)
            the existing index number of the item in the
            deque whose value you want to replace.

        value (any python value/object)
            object/value to replace the current object/value
            stored in the index.
        """
        ### retrieve value using the corresponding original
        ### index from the proxied deque

        orig_index = self.get_original_index(index)
        proxied_val = self.proxied.get_mem_value(orig_index)

        ### if the value retrieved is the same as the one
        ### passed as the argument, then proceed with the
        ### item setting
        if proxied_val is value:
            super().__setitem__(index, value)

        ### otherwise raise an error reporting the
        ### non-compliance of the value
        else:
            raise ValueError("value must be identical to proxied deque value")

    def get_content_view(self):
        """Return a "sibling" content view.

        Overrides WalkingDeque.get_content_view.
        """
        return self.proxied.get_content_view()

    ### custom deepcopy operation

    def __deepcopy__(self, memo):
        """Return a deepcopy of this content view.

        Overrides WalkingDeque.__deepcopy__.

        This method is executed automatically by the
        copy.deepcopy function.

        memo (dict)
            provided automatically by the copy.deepcopy
            function. Used by copy.deepcopy to keep track
            of objects which were already deepcopied, thus
            avoiding deepcopying the same object more than
            once.
        """
        ### get proxied deepcopy reference
        proxied_deepcopy = deepcopy(self.proxied, memo)

        ### instantiate new content view
        new_content_view = proxied_deepcopy.get_content_view()

        ### record (id(self) -> new_content_view) data pair
        ### in the memo mapping
        memo[id(self)] = new_content_view

        return new_content_view

    ### custom string representation

    def __repr__(self):
        """Return an unambiguous string representation.

        Extends WalkingDeque.__repr__.
        """
        return super().__repr__().replace("WalkingDeque", "ContentView", 1)



"""Facility for index sequences generation."""

### standard library imports

from collections import deque
from collections.abc import Sequence
from random import seed, shuffle

### local import
from .wdeque.main import WalkingDeque


# XXX
# idea: implement progressions in indices generation
# that is, geometric progression, etc.
# maybe fibonacci too


def generate_indices(index_data, index_map={}):
    """Return indices list after generating/changing them.

    index_data (dict or list)
        data used to retrieve/load/generate/alter
        indices recursively or not.
    index_map (dict)
        contains indices lists. It is only meant to provide
        values and never be altered here.
    """
    ### if data provided is a sequence, then you don't
    ### need to process anything; just return it
    if isinstance(index_data, Sequence):
        return index_data

    ### process operation to
    ### retrieve/load/generate/alter surfaces

    name = index_data["name"]

    ## just add provided indices
    if name == "retrieve_from_map":
        key = index_data["key"]
        output = index_map[key]

    ## generate indices from range objects
    elif name == "range":
        start = index_data.get("start", 0)
        stop = index_data["stop"]
        step = index_data.get("step", 1)

        output = list(range(start, stop, step))

    ## generate indices by repeating indices
    elif name == "repeat":
        indices = generate_indices(index_data["sequence"], index_map)

        how_many_times = index_data["how_many_times"]

        output = indices * how_many_times

    ## generate indices by executing pre-defined
    ## operation
    elif name == "back_and_forth":
        length = index_data["original_length"]

        first_half = list(range(length))
        last_half = list(range(length)[::-1][1:-1])

        if index_data.get("is_eased", False):
            first_half = generate_indices(
                {"name": "easing", "index_data": first_half}, index_map
            )

            last_half = generate_indices(
                {"name": "easing", "index_data": last_half}, index_map
            )

        output = first_half + last_half

    ## change indices by multiplying existing ones
    elif name == "multiply":
        indices = generate_indices(index_data["sequence"], index_map)

        multiplier = index_data.get("multiplier", False)

        output = []

        ## multiply all indices
        if multiplier:
            for item in indices:
                seq = [
                    item,
                ] * multiplier
                output.extend(seq)

        ## multiply only specified indices
        else:
            multiplier_map = index_data["multiplier_map"]
            for item in indices:
                multiplier = multiplier_map.get(str(item), 1)
                seq = [
                    item,
                ] * multiplier

                output.extend(seq)

    ## change indices by slicing existing ones
    elif name == "slice":
        indices = generate_indices(index_data["index_data"], index_map)

        start = index_data.get("start")
        stop = index_data.get("stop")
        step = index_data.get("step")

        output = indices[start:stop:step]

        # note: collections.deque can't be sliced; the
        # walking deque, which is a subclass probably
        # can't too.

    ## change indices by rotating existing ones using
    ## deque
    elif name == "deque_rotate":
        indices = generate_indices(index_data["index_data"], index_map)

        rotation_amount = index_data["rotation_amount"]

        indices_deque = deque(indices)
        indices_deque.rotate(rotation_amount)

        output = list(indices_deque)

    ## get a walking deque from the indices
    elif name == "get_walking_deque":

        # get indices

        indices = generate_indices(index_data["index_data"], index_map)

        # create walking deque with them
        output = WalkingDeque(indices)

    ## change indices by executing pre-defined operation
    elif name == "easing":
        ## create an easing effect on values in the edges
        indices = generate_indices(index_data["index_data"], index_map)

        length = len(indices)
        output = []

        if length % 2:
            levels = length // 2

            ## iterate from center to first index
            for n, i in enumerate(range(levels, -1, -1), 1):
                seq = [i] * n
                output = seq + output

            ## iterate from center to last index
            for n, i in enumerate(range(levels, length), 1):
                if n == 1:
                    continue
                seq = [i] * n
                output += seq
        else:
            right = length // 2
            left = right - 1

            for n, i in enumerate(range(left, -1, -1), 1):
                seq = [i] * n
                output = seq + output
            for n, i in enumerate(range(right, length), 1):
                seq = [i] * n
                output += seq

        output = list(map(indices.__getitem__, output))

    ## change indices by executing pre-defined operation
    elif name == "shuffle":
        indices = generate_indices(index_data["index_data"], index_map)

        seed()
        shuffle(indices)

        output = indices

    ## concatenate multiple sequences
    elif name == "concatenate":
        output = []
        for sequence_data in index_data["sequences"]:
            output += generate_indices(sequence_data, index_map)

    ### make sure return value is a sequence

    if isinstance(output, Sequence):
        return output

    else:
        return [output]



"""Common utilities for handling tree-structured data."""


def get_tree_values(tree, key_name, children_key_name):
    """Return all values of given key in tree mapping.

    It walks all the tree recursively retrieving the
    value in the key provided.

    tree
        A mapping structured as a tree.
    key_name
        String representing field whose value you want to
        retrieve.
    children_key_name
        String with which to retrieve a list of
        children mapping.
    """
    values = []
    try:
        value = tree[key_name]
    except KeyError:
        pass
    else:
        values.append(value)

    try:
        children_data = tree[children_key_name]
    except KeyError:
        pass
    else:
        for child_data in children_data:
            child_values = get_tree_values(child_data, key_name, children_key_name)
            values.extend(child_values)

    return values


def get_tree_height(tree, height=0):
    """Get the height of the tree.

    tree
        A mapping structured as a tree.
    height
        Integer indicating height of the tree argument.
        When passing the entire tree, it should be 0.

    ### doctests

    >>> dir_tree = {
    ...   "type": "directory",
    ...   "name": "hello",
    ...   "children": [
    ...     {
    ...       "type": "directory",
    ...       "name": "world",
    ...       "children": [
    ...         {"type": "file", "name": "one.txt"},
    ...         {"type": "file", "name": "two.txt"}
    ...       ]
    ...     },
    ...     {"type": "file", "name": "README"}
    ...   ]
    ... }
    >>> get_tree_height(dir_tree)
    2
    >>> dummy_tree = {
    ...   "name": "root",
    ...   "children": [
    ...     {
    ...       "name": "child1",
    ...       "children": [
    ...         {"name": "grandchild1"},
    ...         {
    ...           "name": "grandchild2",
    ...           "children" : [
    ...             {"name" : "grandgrandchild1"}
    ...           ]
    ...         },
    ...         {"name": "grandchild3"}
    ...       ]
    ...     },
    ...     {"name": "child2"}
    ...   ]
    ... }
    >>> get_tree_height(dummy_tree)
    3
    >>> another_tree = {"name" : "root"}
    >>> get_tree_height(another_tree)
    0

    """
    ### verify if tree has children
    try:
        children = tree["children"]

    ### if there are no children
    except KeyError:
        pass

    ### if it has children, increment the height
    ### and pass it on while calculating the height
    ### of each child recursively
    else:
        height += 1

        height = max(get_tree_height(child, height) for child in children)

    ### return final height
    return height



"""Facility for controls configuration.

Here we provide an object which manages the screen.
It's purpose is to present the player with means to
change settings in the USER_SETTING["keys_map"] mapping and have them applied.
"""

from collections import deque
from functools import partialmethod

### third-party imports

from pygame import (
    QUIT,
    KEYDOWN,
    K_ESCAPE,
    K_RETURN,
    K_w,
    K_UP,
    K_s,
    K_DOWN,
    MOUSEBUTTONDOWN,
    MOUSEMOTION,
)

from pygame.key import name as name_key
from pygame.draw import rect as draw_rect
from pygame.event import get as get_events
from pygame.display import update

### local imports

from ...config import SCREEN, SCREEN_RECT, USER_SETTINGS, DEFAULT_SETTINGS, KEYS_MAP

from ...common.behaviour import empty_function

from ...appcommon.surf import render_rect
from ...appcommon.style import give_depth_finish
from ...appcommon.dialog import cache_message, display_message
from ...appcommon.autoblit import BlitterSet
from ...appcommon.exception import QuitGameException, ManagerSwitchException
from ...appcommon.text.main import render_text
from ...appcommon.behaviour.setting import apply_user_settings

from .widgets import get_button

from ...sound import SOUNDS_MAP
from ...palette import CERULEAN, BLACK, WHITE, SKY_COLOR
from ...keymapping import VALID_KEYS_MAP


# TODO
# I should probably use strings for control values instead
# of their integer values; this ways the same controls
# scheme can be used in different systems;


class ControlsWidget:
    """Widget for setting up custom controls.

    The user is allowed to change controls for some
    specific actions.
    """

    def __init__(self, options_widget):
        """Set variables and perform setups.

        options_widget
            The instance of mainmenu.options.OptionsWidget.
        """
        self.options_widget = options_widget

        self.set_labels()

        ### Instantiate and store a background surface
        self.background = render_rect(*SCREEN_RECT.size, color=SKY_COLOR)

        self.build_widgets()
        self.set_other_objects()

        cache_message(
            "not_free", "The key you chose is already being used" + " in another action"
        )
        cache_message("not_valid", "The key you chose isn't considered valid.")

        ### define behaviours

        self.draw = self.draw_control_widgets
        self.control = self.handle_events

    def set_labels(self):
        """Create a sprite to serve as a label."""
        self.labels = BlitterSet()

        self.texts = [
            "Player left movement",
            "Player right movement",
            "Player up movement",
            "Player down movement",
            "Player jump movement",
            "Player fire item action",
            "Player interact action",
            "Advance dialogue/deny prompt",
            "Accept prompt",
        ]
        y = 25
        self.max_width = 0
        for text in self.texts:
            label = render_text(
                text,
                32,
                foreground_color=WHITE,
                background_color=BLACK,
                return_obj=True,
                coordinates_value=(60, y),
            )
            label.text = text
            y += 60
            self.labels.add(label)

            self.max_width = max(self.max_width, label.rect.w)

        for label in self.labels:
            label.rect.x = label.rect.x + self.max_width - label.rect.w

    def build_widgets(self):
        """Create option widgets."""
        self.widgets_deque = deque()

        fields = [
            "player_left",
            "player_right",
            "player_up",
            "player_down",
            "jump",
            "fire_item",
            "interact",
            "advance_deny",
            "accept_prompt",
        ]

        arbitrary_offset = 80
        x = self.max_width + arbitrary_offset
        y = 25
        for text, field in zip(self.texts, fields):
            widget = ControlSwitcher(self, field, (x, y))
            widget.text = text
            self.widgets_deque.append(widget)
            y += 60
        self.widgets_deque.append(
            get_button("Restore default controls", (x, y), self.restore_defaults)
        )

        self.widgets = BlitterSet(self.widgets_deque)

        self.current_widget = self.widgets_deque[0]
        self.highlight_rect = self.current_widget.rect

    def set_other_objects(self):
        """Create additional objects to help."""
        self.prompt_labels = BlitterSet()

        txt01 = "Please, type the new key"
        txt02 = "for the action you chose:"
        txt03 = "(or press <ESCAPE> if you want to cancel)"
        type_prompt01 = render_text(
            txt01,
            font_size=42,
            foreground_color=WHITE,
            background_color=BLACK,
            return_obj=True,
        )

        type_prompt02 = render_text(
            txt02,
            font_size=42,
            foreground_color=WHITE,
            background_color=BLACK,
            return_obj=True,
        )
        type_prompt03 = render_text(
            txt03,
            font_size=28,
            foreground_color=WHITE,
            background_color=BLACK,
            return_obj=True,
        )

        type_prompt01.rect.midbottom = type_prompt02.rect.midtop = SCREEN_RECT.center

        self.action_name_topleft = type_prompt02.rect.bottomleft

        type_prompt03.rect.bottomright = SCREEN_RECT.bottomright
        type_prompt03.rect.x += -35
        type_prompt03.rect.y += -35

        self.prompt_labels.update([type_prompt01, type_prompt02, type_prompt03])

    def handle_events(self):
        """Update widget state based on event_queue."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            elif event.type == KEYDOWN:

                if event.key == K_ESCAPE:
                    raise ManagerSwitchException(self.options_widget)

                elif event.key == K_RETURN:
                    self.invoke_button()

                elif event.key in (K_w, K_UP):
                    self.go_up()

                elif event.key in (K_s, K_DOWN):
                    self.go_down()

            elif event.type == MOUSEMOTION:
                self.mouse_motion_routine(event.pos)

            elif event.type == MOUSEBUTTONDOWN:
                if event.button == 1:
                    self.mouse_click_routine(event.pos)

    def update(self):
        """Update widgets."""
        self.widgets.update_objs()

    def draw_control_widgets(self):
        """Draw control widgets and other widgets."""
        SCREEN.blit(self.background, (0, 0))
        self.labels.draw()
        self.widgets.draw_objs()
        draw_rect(SCREEN, CERULEAN, self.highlight_rect, 4)

        update()

    def change_field_key(self, field):
        """Change the field key input."""
        self.field = field

        self.current_label = [
            label for label in self.labels if self.current_widget.text == label.text
        ].pop()

        self.draw = self.show_key_prompt
        self.control = self.get_new_key

    def show_key_prompt(self):
        """Prompt user to type new key."""
        SCREEN.blit(self.background, (0, 0))
        self.prompt_labels.draw()
        SCREEN.blit(self.current_label.image, self.action_name_topleft)

        update()

    def get_new_key(self):
        """Get new key typed by user."""
        for event in get_events():
            if event.type == QUIT:
                raise QuitGameException

            elif event.type == KEYDOWN:

                if event.key == K_ESCAPE:

                    self.draw = self.draw_control_widgets
                    self.control = self.handle_events

                else:
                    self.check_key(event.key)

    def check_key(self, key):
        """Verify and switch key if appropriate."""
        valid = key in VALID_KEYS_MAP.values()
        free = key not in KEYS_MAP.values()

        if free and valid:

            KEYS_MAP[self.field] = key
            self.current_widget.update_surface()

            apply_user_settings()

            self.draw = self.draw_control_widgets
            self.control = self.handle_events

        elif not free:
            SOUNDS_MAP["error"].play()
            display_message("not_free", self)

        elif not valid:
            SOUNDS_MAP["error"].play()
            display_message("not_valid", self)

    def switch_widget(self, steps=0):
        """Switch widgets by steps.

        steps
            integers representing how many the widgets deque
            should rotate.
        """
        self.widgets_deque.rotate(steps)
        self.current_widget = self.widgets_deque[0]

        self.highlight_rect = self.current_widget.rect
        SOUNDS_MAP["gui_step"].play()

    go_up = partialmethod(switch_widget, 1)
    go_down = partialmethod(switch_widget, -1)

    def mouse_motion_routine(self, mouse_position):
        """Check if mouse touches any widget and select it."""
        touched_item = None

        for item in self.widgets_deque:

            if item.rect.collidepoint(mouse_position):
                touched_item = item
                break

        if touched_item == self.current_widget:
            return

        if touched_item:
            for i in range(len(self.widgets_deque)):

                if touched_item == self.widgets_deque[0]:
                    self.switch_widget()
                    break
                self.widgets_deque.rotate(1)

    def mouse_click_routine(self, mouse_position):
        """Invoke widget if mouse click over it."""
        touched_item = None

        for item in self.widgets_deque:

            if item.rect.collidepoint(mouse_position):
                touched_item = item
                break

        if touched_item:

            if touched_item != self.current_widget:
                self.mouse_motion_routine(mouse_position)

            touched_item.invoke()

    def restore_defaults(self):
        """Restore control defaults."""
        USER_SETTINGS["keys_map"].update(DEFAULT_SETTINGS["keys_map"])

        for widget in self.widgets:

            try:
                widget.update_surface()
            except AttributeError:
                pass

        apply_user_settings()

    def invoke_button(self):
        """Invoke current widget, if it can be invoked."""
        self.current_widget.invoke()


class ControlSwitcher:
    """Widget to change control for specific action."""

    def __init__(self, controls_widget, field, topleft):
        """Assign data and perform setups."""
        self.controls_widget = controls_widget

        self.field = field
        self.topleft = topleft

        self.update_surface()

        ### define update behaviour
        self.update = empty_function

    def draw(self):
        """Update object state."""
        SCREEN.blit(self.image, self.rect)

    def invoke(self):
        """Change key of the action."""
        self.controls_widget.change_field_key(self.field)

    def update_surface(self):
        """Update image surface."""
        text = name_key(KEYS_MAP[self.field])

        if text == "return":
            text = "enter"

        surf = render_text(
            text, font_size=32, foreground_color=WHITE, background_color=BLACK
        )

        self.image = give_depth_finish(surf)

        self.rect = self.image.get_rect()
        self.rect.topleft = self.topleft



"""Facility for time related custom pygame tools."""

from ..logconfig import APP_LOGGER
from ..config import MILLISECS_PER_FRAME


logger = APP_LOGGER.getChild(__name__)


class TaskManager(object):
    """Manages task objects between differents levels."""

    def __init__(self):
        """Assign variables."""
        self.tasks = []
        self.sendoff_tasks = []

    def clear(self):
        """Set defaults and perform setups."""
        self.tasks.clear()

        for task in self.sendoff_tasks:
            try:
                task()
            except Exception as err:
                err_msg = str(err)
                print(err_msg)
                logger.exception(err_msg)

        self.sendoff_tasks.clear()

    def update(self):
        """Update tasks."""
        finished_tasks = []
        for task in self.tasks:
            try:
                task.update()
            except Exception as err:
                err_msg = str(err)
                print(err_msg)
                logger.exception(err_msg)

                task.finished = True

            if task.finished:
                finished_tasks.append(task)

        for task in finished_tasks:
            self.tasks.remove(task)

    def add_task(self, callable_task, delta_t, unit="milliseconds", cyclic=False):
        """Instantiate and store task. Also return it.

        callable_task
            Callable to be executed when delta_t is reached.
        delta_t
            Milliseconds or frames until execution of
            callable.
        unit
            String hinting if time interval (delta_t) is
            measured in milliseconds or frames.
            Defaults to 'milliseconds'.
        cyclic
            Boolean. If True, the task will execute
            indefinitely, until either its "finished"
            attribute is manually changed to True or
            the task is cleared off the task manager's
            'tasks' attribute. Defaults to False.
        """
        task = TimedTask(callable_task, delta_t, unit, cyclic)
        self.tasks.append(task)

        return task

    def add_sendoff_task(self, callable_task):
        """Store a callable to call before leaving level.

        callable_task
            A callable to be called before exiting a level.
            Callables are executed in the self.clear method.
        """
        self.sendoff_tasks.append(callable_task)


TASK_MANAGER = TaskManager()
add_task = TASK_MANAGER.add_task
add_sendoff_task = TASK_MANAGER.add_sendoff_task
clear_task_manager = TASK_MANAGER.clear
update_task_manager = TASK_MANAGER.update


class TimedTask(object):
    """A task which executes after a time interval.

    Tasks are scheduled either in milliseconds or in frames.
    They can be executed only once or repeatedly.
    """

    # XXX
    # Instead of changing the update method,
    # just change the line where time is incremented
    # by creating a self.increment attribute
    # which is either MILLISECS_PER_FRAME or 1.
    #
    # Also do the same for the execute method, just
    # isolate the line where the finished state is
    # determined.
    #
    # Also, for unit selections use elif unit == "frames";
    # the else option should raise and error

    def __init__(self, callable_task, delta_t, unit="milliseconds", cyclic=False):
        """Assign variables.

        callable_task
            Callable to be executed when delta_t is reached.
        delta_t
            Time interval measured in milliseconds or frames.
        unit
            String hinting if time interval (delta_t) is
            measured in milliseconds or frames.
            Defaults to 'milliseconds'.
        cyclic
            Boolean. If True, the task will execute
            indefinitely, until either its "finished"
            attribute is manually changed to True or
            the level ends. Defaults to False.
        """
        self.finished = False

        self.delta_t = delta_t
        self.invoke = callable_task

        self.start = 0

        if unit == "milliseconds":
            self.update = self.count_milliseconds
        else:
            self.update = self.count_frames

        if cyclic:
            self.execute = self.execute_cycle
        else:
            self.execute = self.execute_once

    def count_milliseconds(self):
        """Count milliseconds to execute task."""
        if self.start >= self.delta_t:
            self.execute()

        self.start += MILLISECS_PER_FRAME

    def count_frames(self):
        """Count frames to execute task."""
        if self.start >= self.delta_t:
            self.execute()

        self.start += 1

    def get_remaining(self):
        """Get remaining time."""
        if not self.finished:
            return self.delta_t - round(self.start)

        else:
            return 0

    def execute_once(self):
        """Execute and finish task."""
        self.invoke()
        self.finished = True

    def execute_cycle(self):
        """Execute and restart another cycle."""
        self.invoke()
        self.start = 0


# XXX
# Future tasks which executes based on conditions
# could be cyclic too. In fact, depending on the
# design it could be possible to generate a great
# amount of possibilities by combining the parameters
# of each kind of task. Ponder.
